! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!
! Module: bkgoptrate_aux
!
! Authors:  Masaaki Sakano (SSC/LUX)
!          
! Description: Core subroutines/functions for bkgoptrate.
! 
! Comments: 
! 
! TODO: 
!  * Clean up findOptimumY(Strict|StrictClean|Crude).  At the moment Strict is used, but its algorithm is messy, as it is just an update of Crude...  StrictClean is meant to be clean, but it does not pass the test harness.  Even cleaner algorithm is well possible.
!
! $Id: bkgoptrate_aux.f90,v 1.5 1998/01/13 08:39:15 gvacanti Exp $
!- 
module bkgoptrate_aux

  use dal	! For type definition of single etc.
  ! use param
  use errorHandling
  ! use datafilter

  ! use array_utils	! for smallest/largestTrueIndex
  ! use ssc_misc		! for isNaN

  implicit none 

  private 

  public initialiseBkgoptrateDiagnosisT
  public getSN2ofTs
  public getSN2ofTsOfRate
  public findOptimumY
  public calcThresholdFromAny
  public calcThresholdFromRate
  public calcThresholdFromCount

  type, public :: InfoFracexpT
    character(20) :: style = ''	! (calc|threshold|none)
      ! calc: Calculate the count or rate according to the FRACEXP
      ! threshold: Ignore the bins with FRACEXP lower than the threshold
    logical :: doUse = .false.
    character(64) :: colName = ''	! 'FRACEXP'
    real(single)  :: lowerThreshold = 0.0	! valid if style=='threshold'
    real(double), allocatable :: ary(:)
  end type InfoFracexpT

  type, public :: ArgTimesStartEndT
    logical :: withStart=.false., withEnd=.false.
    real(single)  :: start=0.0, end=0.0
  end type ArgTimesStartEndT

  type, public :: BkgoptrateDiagnosisT
    real(single), allocatable :: aryBRawCut(:), aryBRateCut(:), arySN(:), aryRate(:)
    integer,      allocatable :: aryNBin(:)
  end type BkgoptrateDiagnosisT

  ! interface getSN2ofTs
  !   module procedure getSN2ofTsWithoutExp, getSN2ofTsWithExp
  ! end interface

contains 

  !---------------------------------------------------------------------
  subroutine initialiseBkgoptrateDiagnosisT(arySize, bkgDiagn)
  !---------------------------------------------------------------------
  ! Returns the square of S/N for the given time-series.

    implicit none

    integer, intent(in) :: arySize
    type(BkgoptrateDiagnosisT), intent(inout) :: bkgDiagn

    !-------------------------------------------------------------------
    ! Start

    allocate(bkgDiagn%aryBRawCut(arySize) &
           , bkgDiagn%aryBRateCut(arySize) &
           , bkgDiagn%aryRate(arySize)  &
           , bkgDiagn%arySN(arySize)  &
           , bkgDiagn%aryNBin(arySize))

  end subroutine initialiseBkgoptrateDiagnosisT


  !---------------------------------------------------------------------
  function getSN2ofTs(tSeries, mask, aryExp) result(sn2)
  !---------------------------------------------------------------------
  ! Returns the square of S/N for the given time-series (of Count),
  ! for a source with no time variability.
  ! Optional aryExp is the array of FRACEXP, therefore (0.0 <= aryExp <= 1.0).
  ! Strictly speaking, the true S/N is (SourceCount * getSN2ofTs()).
  ! Note: Returns 0.0, if the (mask && aryExp) turns out to be all False.

    implicit none

    real(single) :: sn2
    real(single), intent(in) :: tSeries(:)
    logical,      intent(in), optional :: mask(size(tSeries))
    real(single), intent(in), optional :: aryExp(size(tSeries))

    logical :: maskLoc(size(tSeries))
    real(single) :: denominator
! logical, parameter :: localIsTest = .true.
logical, parameter :: localIsTest = .false.

    !-------------------------------------------------------------------
    ! Start

    denominator = 0.0	! Initialisation

    if (present(aryExp)) then
      if (present(mask)) then
        maskLoc = mask
if (localIsTest) print *,'DEBUG(getSN2ofTs):mask(withAryExp)=',mask
      else
        maskLoc = .true.
      end if
if (localIsTest) print *,'DEBUG(getSN2ofTs):maskLoc(withAryExp)1=',maskLoc
      where(aryExp <= 0.0)
        maskLoc = .false.	! Bins, where aryExp==0.0, are ignored.
      end where
if (localIsTest) print *,'DEBUG(getSN2ofTs):maskLoc(withAryExp)2=',maskLoc

      if (any(maskLoc)) then
        ! Do nothing
      else	! All .false.
        sn2 = 0.0
        return
        ! sn2 = sum(aryExp, MASK=maskLoc) / denominator	! This would cause "Runtime Error: *** Arithmetic exception: Floating invalid operation - aborting", but not NaN.
      end if
      ! print *,'DEBUG:maskLoc=',maskLoc
      ! print *,'DEBUG:aryExp=',aryExp,' sum(aryExp)=',sum(aryExp, MASK=maskLoc)**2
      ! print *,'DEBUG:tSeries=',tSeries,' sum(tSeries)=',sum(tSeries, MASK=maskLoc)

      denominator = sum(tSeries, MASK=maskLoc)
      sn2 = sum(aryExp, MASK=maskLoc)**2  / denominator
      ! sn2 = sum(aryExp, MASK=maskLoc)**2  / sum(tSeries, MASK=maskLoc)
if (localIsTest) print *,'DEBUG(getSN2ofTs):sn2(1)=',sn2
      ! if (isNaN(sn2)) then	! defined in ssc_misc
      !   if (localIsTest) then
      !     print *,'DEBUG(NaN):maskLoc=',maskLoc
      !     print *,'DEBUG(NaN):aryExp=',aryExp,' sum(aryExp)=',sum(aryExp, MASK=maskLoc)**2
      !     print *,'DEBUG(NaN):tSeries=',tSeries,' sum(tSeries)=',sum(tSeries, MASK=maskLoc)
      !   end if
      ! end if
    else
      if (present(mask)) then
        if (any(mask)) then
          ! Do nothing
        else	! All .false.
          sn2 = 0.0
          return
          ! sn2 = sum(tSeries, MASK=mask)   / sum(tSeries, MASK=mask)	! This would cause "Runtime Error: *** Arithmetic exception: Floating invalid operation - aborting", but not NaN.
        end if

if (localIsTest) print *,'DEBUG(getSN2ofTs):mask(withoutAryExp)=',mask
        sn2 = real(count(mask))**2   / sum(tSeries, MASK=mask)	! real() is needed to avoid a potential overflow in Integer --- anyhow conversion to Real would be internally needed to give the Real result to match with the 2nd clause (sum of Real).
if (localIsTest) print *,'DEBUG(getSN2ofTs):sn2(2)=',sn2
      else
        sn2 = real(size(tSeries))**2 / sum(tSeries)
if (localIsTest) print *,'DEBUG(getSN2ofTs):sn2(3)=',sn2
      end if
    end if

! sn2 = getSN2ofTs_bak(tSeries, mask, aryExp)	! If this is commented out, the test harness of testbkgoptrateaux01 (and testbkgoptratetask4, if the section of "Mask out the bins with FRACEXP==0" in findOptimumYStrict() was commented in) should fail.

  end function getSN2ofTs


  !---------------------------------------------------------------------
  function getSN2ofTsOfRate(tSeries, mask, aryExp) result(sn2)
  !---------------------------------------------------------------------
  ! Returns the square of S/N for the given time-series of Rate.
  ! 
  ! NOTE: It is probably better to convert the time-series of rate
  !  to that of count beforehand and call getSN2ofTs() instead,
  !  considering the overhead to call this routine repeatedly.
  !   For that reason, this subroutine has not been tested!  Treat with care...

    implicit none

    real(single) :: sn2
    real(single), intent(in) :: tSeries(:)
    logical,      intent(in), optional :: mask(size(tSeries))
    real(single), intent(in), optional :: aryExp(size(tSeries))

    real(single) :: localAryExp(size(tSeries)), denominator
    integer(int32) :: i

    !-------------------------------------------------------------------
    ! Start

    localAryExp = aryExp
    denominator = 0.0
    if (present(mask)) then
      where(.not. mask) localAryExp = 0.0	! Propagates the input mask.
    end if

    do i=1, size(tSeries)
      denominator = denominator + localAryExp(i) * tSeries(i)
    end do

    sn2 = sum(localAryExp)**2 / denominator
    ! NOTE: I am not sure if NaN is treated all right without causing an error.

  end function getSN2ofTsOfRate


  !---------------------------------------------------------------------
  subroutine findOptimumY(tsCnt, optimumY &
     , optimumSize, tsRate, aryFracexp, mask, lowerNbinRatio, status &
     , bkgDiagn, doDump)
  !---------------------------------------------------------------------

    implicit none

    real(single),   intent(in) :: tsCnt(:)
    real(single),   intent(out) :: optimumY
    integer(int32), intent(out), optional :: optimumSize
    real(single),   intent(in),  optional :: tsRate(size(tsCnt)), aryFracexp(size(tsCnt))
    ! logical,        intent(in),  optional :: isRate	! This has to be provided if aryExp(:) is given.
    logical,        intent(inout), optional :: mask(size(tsCnt))
    ! logical,        intent(out), optional :: maskOut(size(tsCnt))
    real(single),   intent(in),  optional :: lowerNbinRatio
    integer,        intent(out), optional :: status
    type(BkgoptrateDiagnosisT), intent(out), optional :: bkgDiagn
    logical,        intent(in),  optional :: doDump

logical, parameter :: localIsTest = .false.

    !-------------------------------------------------------------------
    ! Start

    ! call findOptimumYCrude(tsCnt, optimumY &
    ! call findOptimumYStrictClean(tsCnt, optimumY &
    call findOptimumYStrict(tsCnt, optimumY &
       , optimumSize, tsRate, aryFracexp, mask, lowerNbinRatio, status &
       , bkgDiagn, doDump)

  end subroutine findOptimumY


  !---------------------------------------------------------------------
  subroutine findOptimumYStrict(tsCnt, optimumY &
     , optimumSize, tsRate, aryFracexp, mask, lowerNbinRatio, status &
     , bkgDiagn, doDump)
  !---------------------------------------------------------------------
  ! *** Description ***
  !  Find the optimum cut of Y-value (Count or Rate) that gives the maximum
  ! S/N ratio for the input time-series, which should be the background
  ! time-series, where
  !    S/N =~ Tot_EXP / SQRT(Tot_Cnt)  (x Rate (=Constant) to give the counts).
  !  Then you (the calling routine) can exclude those high-background periods
  ! from another source time-series, based on the optimum background cut.
  !
  ! *** Arguments ***
  !  tsCnt(:) : Array of Counts for time bin.
  !    This is used to estimate the S/N.
  !    Note the unit has to be count (or alike).  Threrefore, if the original
  !    time-series is in unit of Rate, tsCnt(:) has to be calculated and
  !    given to this subroutine by the caller.
  !  optimumY   : The calculated optimum Y-value.
  !  optimumSize : The size of tsCnt(:) when it gives the best S/N (=optimumY).
  !    This is equivalent to count(mask) (after mask is modified in this routine).
  !  tsRate(:) : Array of Rate for each time bin (OPTIONAL).  If not given,
  !    (1) if aryFracexp is not given either, tsRate is set to be tsCnt,
  !    (2) if aryFracexp given, it is calculated accordingly.
  !    tsRate and aryFracexp should not be independent, but this subroutine
  !    does not check consistency when both the arguments are given.
  !    This is used to estimate the optimum threshold.
  !  aryFracexp(:) : Array of FRACEXP (OPTIONAL).  If not given,
  !    (1) if tsRate is not given either, aryFracexp(:) is set to be all 1.0,
  !    (2) if tsRate given, it is calculated accordingly.
  !    tsRate and aryFracexp should not be independent, but this subroutine
  !    does not check consistency when both the arguments are given.
  !  mask(:)    : Mask Array, indicating which indices in tsCnt(:) are valid.
  !    This mask is modified in the subroutine, and the returned mask shows
  !    true for all the valid values when the optimum value is found.
  !  lowerNbinRatio : Threshold in ratio for the number of bin used,
  !    where the calculation stops.  For example, if it is 0.05,
  !    this routine does not search for the optimum Y-value beyond the point,
  !    where the number of bins used is 5% of the entire bin number
  !    after the given mask is applied; that is, for example,
  !    when the size(tsCnt) is 1000 and the given mask has 80% of .true.,
  !    this routine does not search for the optimum value for the number
  !    of bins used being less than 40 (=1000*0.8*0.05), and if it happens,
  !    either error('noThresholdFound') is raised or status=-2 is returned.
  !    The default is 0.0.
  !  bkgDiagn	: type(BkgoptrateDiagnosisT), gives the diagnostic arrays,
  !    the size of which is equal to or likely smaller than tsCnt(:),
  !    because any duplication in tsCnt(:) is eliminated.
  !  status     : Result (fitting) status indicator (OPTIONAL).
  !    0 for normal end,
  !    -2 if the optimum values is not found even when the search reaches the end
  !    (in that case, if status is not given, Error is raised.).
  !    -3 if in addition to the above case of -2, search has completely failed.
  !    -4 if the given mask has no True.
  !  doDump     : For debugging - it dumps all the S/Ns if True.
  !    Note the returned values are most likely to be incorrect if True.
  !
  ! *** ALGORITHM ***
  !   0. Prepare TSeries(:), TRate(:) and Mask(:).
  !   1. Calculate the S/N in TSeries, based on Mask.
  !   2. Find the maximum value and its locations, namely the indices, in TRate(:).
  !   3. Set Mask(:) false for the index(ices), which have the maximum value,
  !  and calculate the S/N in TSeries, based on (revised) Mask.
  !   4. If the new S/N is larger than the previous S/N,
  !  then repeat the processing from (2).  If smaller,
  !  then the previous maximum value and Mask are the candidate optimum ones
  !  (best S/N).  Repeat processing from (2) till all the S/N is calculated,
  !  and at last return the parameter that gives the maximum S/N.
  !
  ! *** TODO ***
  !  * At the time of writing, the algorithm to search for the parameter
  ! that gives the maximum S/N, is based on the minimization (maximization)
  ! algorithm, that is, it tries to find local maximums step by step.
  ! However, it is unnecessarily complex.  It would be far better
  ! to simply calculate all the possible S/N first and find the maximum.
  ! Only the trouble to do it is to calculate the corresponding mask file.
  !
  !  * It would be simpler if all the input time-series are sorted
  ! in order of Count (or Rate or corrected Rate with FRACEXP,
  ! depending on the requirement) before they are processed.
  ! Perhaps use ssclib/sort_mod for it.
  !

    implicit none

    real(single),   intent(in) :: tsCnt(:)
    real(single),   intent(out) :: optimumY
    integer(int32), intent(out), optional :: optimumSize
    real(single),   intent(in),  optional :: tsRate(size(tsCnt)), aryFracexp(size(tsCnt))
    logical,        intent(inout), optional :: mask(size(tsCnt))
    real(single),   intent(in),  optional :: lowerNbinRatio
    integer,        intent(out), optional :: status
    type(BkgoptrateDiagnosisT), intent(out), optional :: bkgDiagn
    logical,        intent(in),  optional :: doDump

    integer(int32) :: optimumSizeLocMax	! optimumY for the last local max S/N
    logical, dimension(size(tsCnt)) :: prevMask, curMask, locMaxMask
    logical :: isFracexpUsed, localDoDump
    real(single), dimension(size(tsCnt)) :: localTsRate &
       , aryExp, aryBRawCut, aryBRateCut, arySN, aryRate	! for ary*(:), the size is the maximum possible, but the real size needed is likely to turn out to be smaller.
    integer,      dimension(size(tsCnt)) :: aryNBin
    real(single) :: prevSN, curSN, bestSN, prevprevY, prevY, curY	! prevprevY can now be deleted...?
    integer(int32) :: nbin, sizeMaskIn, tmpSize, lowerNbin, nDup, iDiag
    integer :: arMaxloc(1), prevLoc, curLoc

! logical, parameter :: localIsTest = .true.
logical, parameter :: localIsTest = .false.

    !-------------------------------------------------------------------
    ! Start

    ! *** Algorithm (see also the above comment for the overview) ***
    !
    !  (1) Set False at the X (prevLoc) for the maximum Y (Rate) in the current mask (curMask).
    !  (2) Find the index X for the maximum Y (Rate) with the new mask (curLoc),
    !    as well as the maximum Y-value (curYRate).
    !  (3a) If curY==prevY then repeat from (1) after resetting prevLoc as curLoc.
    !  (3b) If not, proceed below.
    !  (4) Calculate S/N (curSN) for TS with the new mask (curMask).
    !  (5a) If (curSN > prevSN) then repeat from (1) after resetting prev*.
    !  (5b) If not, compare prev* with past best*, and if it gives the best result, best* is replaced with prev* as the candidate best values, and repeat from (1) after resetting prev*.
    !  (6a) If best* is found (nb. the returned Y-value is slightly larger
    !    than prevYRate), store them as the candidate best values,
    !    and repeat from (1) after resetting prev*.
    !  (6b) If fails to find the maximum for SN, then it returns status=-2 or raises an error.
    !    However, even in that case, if over one-third of the lowest bins
    !    have the same Y(Rate)-value, AND if the lowerNbinRatio given is the minimum,
    !    then the Y(Rate)-value (or strictly speaking a value slightly higher
    !    than that) is returned and this subroutines ends normally.
    !    This case may happen in, such as, uniform time-series or step-function-like ones.
    !
    ! *** Notation ***
    !  (prev|cur)Mask : The mask file used to calculate the values.
    !  (prev|cur)SN   : Previous/Current SN when the above mask is applied.
    !  bestSN         : Best SN found so far during the loop.
    !  (prev|cur)Y    : Maximum Y(Rate)-value (when the above mask is applied).
    !  (prev|cur)Loc  : Its index             (when the above mask is applied).
    !  prevprevY      : Previous previous maximum Y(Rate)-value.

    ! isOptimumFound = .false.

    curMask = .true.
! print *, 'DEBUG:curMask=',curMask
    if (present(mask)) curMask = mask
! print *, 'DEBUG:curMask=',curMask

    optimumY = -1
    localTsRate(:) = tsCnt
    aryExp(:) = 1.0
    isFracexpUsed = .false.
    if (present(status)) status = 0
    if (present(optimumSize)) optimumSize = -1
    if (present(tsRate)) then	! Array of Rate is given.
      isFracexpUsed = .true.
      localTsRate = tsRate
      if (present(aryFracexp)) then
        aryExp(:) = aryFracexp
      else
        ! Now, calculate FRACKEXP (=aryExp(:)), based on tsRate and tsCnt.
        where(localTsRate .ne. 0) aryExp(:) = tsCnt/localTsRate
        ! It is impossible to guess aryExp() when tsRate (tsCnt) is zero.
      end if
    else if (present(aryFracexp)) then	! Array of Rate is now calculated.
      isFracexpUsed = .true.
        aryExp(:) = aryFracexp
        where(aryExp == 0)
          localTsRate = 0.0
          ! localTsRate is set to zero where FRACEXP is zero, because to set it to zero is handy for the calculation later in this code.
        elsewhere
          localTsRate = tsCnt/aryExp
        end where
    end if

    sizeMaskIn = count(curMask)
    if (isFracexpUsed) then
      ! Mask out the bins with FRACEXP==0.
      where(aryExp <= 0.0)
        curMask = .false.	! Bins, where aryExp==0.0, are ignored.
      end where

      tmpSize = count(curMask)
      if (sizeMaskIn > tmpSize) then
        write(errStr, '(a,i8,a,i8,a)') 'Some bins have zero FRACEXP &
           &(either given or calculated), which will not be used here, &
           &hence the size of the mask, namely the number of bins used, &
           &is reduced from (', sizeMaskIn, ') to (', tmpSize, ').'
          call message(AppMsg, Sparse)
      end if
      sizeMaskIn = tmpSize
    end if		! if (isFracexpUsed) then

    prevMask = curMask

    if (sizeMaskIn <= 0) then
      write(errStr, '(a)') 'findOptimumY: There is no True in the given mask.'
      if (present(status)) then
        status = -4
        call warning('allMaskFalse') 
        return
      else
        call error('allMaskFalse') 
      end if
    end if	! if (sizeMaskIn <= 0) then

    if (isFracexpUsed) then
      ! prevSN = getSN2ofTs(tsCnt, aryExp, .false., prevMask)
      prevSN = getSN2ofTs(tsCnt, prevMask, aryExp)
    else
      prevSN = getSN2ofTs(tsCnt, prevMask)
      ! prevSN = sizeMaskIn / sqrt(sum(tsCnt, MASK=prevMask))	! S/N calculation.
    end if
    bestSN = -9999.9
    curSN  = -9999.9

    localDoDump = .false.
    if (present(bkgDiagn)) then
      localDoDump = .true.
    else if (present(doDump)) then
      localDoDump = doDump
    end if

    if (present(lowerNbinRatio)) then
      lowerNbin = nint(sizeMaskIn * lowerNbinRatio)
      if ((lowerNbin == 0) .or. (lowerNbin == 1)) then
        lowerNbin = 2	! when lowerNbinRatio=0.0
      else if (lowerNbin < 0) then
        ! lowerNbinRatio has to be zero or positive.
        call error('Bug_lowerNbinRatio', 'Contact the code developer.')
      end if
    else
      lowerNbin = 2
    end if

    if (verbosity(AppMsg) >= Noisy) then
      if (present(lowerNbinRatio)) then
        write(errStr, '(a,i8,a,i8,a,i6,a,f7.4,a)') "Start findOptimumY(): Size TS(" &
           , size(tsCnt), "), mask(", sizeMaskIn &
           , "), lowerNbin=", lowerNbin, " (lowerNbinRatio=", lowerNbinRatio &
           , "  which will be applied further)."
      else
        write(errStr, '(a,i8,a,i8,a,i6,a)') "Start findOptimumY():  Size TS(" &
           , size(tsCnt), "), mask(", sizeMaskIn &
           , "), lowerNbin=", lowerNbin, " (lowerNbinRatio=UNSPECIFIED)."
      end if
      call message(AppMsg, Noisy)
    end if

    if (localIsTest) then
      print *,"DEBUG: nbin=",count(prevMask),"curSN2=",prevSN,"in Start."
      ! print *,"DEBUG: nbin=",count(curMask),"curSN2=",prevSN,"in Start."
    end if

! aryNBin(:) = 0
! print *, 'DEBUG:aryNBin=',aryNBin
! arMaxloc = maxloc(aryNBin)
! print *, 'DEBUG:arMaxloc(aryNBin)=',arMaxloc
! print *, 'DEBUG:prevMask=',prevMask
! print *, 'DEBUG:localTsRate=',localTsRate
    arMaxloc = maxloc(localTsRate, prevMask)
! print *, 'DEBUG:arMaxloc(localTsRate)=',arMaxloc
    prevLoc = arMaxloc(1)
    prevY = localTsRate(prevLoc)	! == maxval(localTsRate, prevMask)
    ! prevY = prevY + abs(epsilon(prevY)) * 4	! slightly larger than the first maximum value.
    prevprevY = prevY 
    nDup = 1

    aryBRawCut(:)  = -1.0	! Initial value.
    aryBRateCut(:) = -1.0
    aryRate(:) = -1.0
    arySN(:)   = -1.0
    aryNBin(:) = -1
    iDiag = 1
    aryBRawCut(iDiag) = tsCnt(prevLoc) + 100*epsilon(tsCnt(prevLoc))	! In count.
    aryBRateCut(iDiag) = prevY + 100*epsilon(prevY)	! 100 is an arbitrary number.
    aryRate(iDiag) = prevY
    arySN(iDiag)   = prevSN
    aryNBin(iDiag) = count(prevMask)

if (localIsTest) print *,'DEBUG: localTsRate=',localTsRate
    do nbin=sizeMaskIn, lowerNbin, -1	! (lowerNbin >= 2) is guaranteed.
      ! nbin is the number of bins to be used in the estimate in each cycle.

      curMask(prevLoc) = .false.
      arMaxloc = maxloc(localTsRate, curMask)	! To calculate curLoc.
      curLoc = arMaxloc(1)
      curY = localTsRate(curLoc)	! Maximum Y for curMask

      if (prevY == curY) then	! The current maximum Y-value is identical to the previous Y-value, so should be taken away before the next evaluation.
        prevLoc = curLoc 
        nDup = nDup + 1
        if (localIsTest) then
if (localIsTest) print *,"DEBUG: nbin=",count(curMask),"                    Removed-(i,Y)=",prevLoc,prevY
        end if
        cycle
      end if

      if (isFracexpUsed) then
        curSN = getSN2ofTs(tsCnt, curMask, aryExp)
        ! NOTE: This formula can replace the one in else-clause, in other words,
        !  this IF clause itself is not essential, however it would be slower.
      else
        curSN = getSN2ofTs(tsCnt, curMask)
      end if

      iDiag = iDiag + 1
      if (aryBRawCut(iDiag) > 0) call error('bugAryBcut','Contact the code developer.')	! This should not happen.
      aryBRawCut(iDiag) = tsCnt(curLoc) + 2*epsilon(tsCnt(curLoc))
      aryBRateCut(iDiag) = min(curY+100*epsilon(curY), (curY+prevY)/2.0)
      aryRate(iDiag) = curY
      arySN(iDiag)   = curSN
      aryNBin(iDiag) = count(curMask)

      !THIS NOTE IS INCORRECT!!!!!
      ! NOTE: Here, count(curMask) should be equal to (nbin-1).
      !  Therefore the following equation should do the same job,
      !  and would be faster due to no overhead.
      ! curSN = (nbin-1) / sqrt(sum(tsCnt, MASK=curMask))
      !
      !  But currently I use the above function just to be conservative and safe.

      if (localIsTest) then
        print *,"DEBUG: nbin=",count(curMask),"curSN2=",curSN,"Removed-(i,Y)=",prevLoc,prevY
      end if

      ! if ((isOptimumFound) .or. (curSN > prevSN)) then
      !   nDup = 1
      !   prevSN = curSN
      !   prevprevY = prevY 
      !   prevY  = curY
      !   prevMask = curMask
      !   prevLoc = curLoc
      !   cycle	! At the time of writing, this is redundant, but just to play safe.
      ! 
      ! else
      if ((curSN < prevSN) .and. (bestSN < prevSN)) then
        ! The previous attempt gives the better S/N than the current and former best,
        ! so it was the local maximum S/N value and candidate best value.

        ! if (isOptimumFound) call error('bugOptimum','Contact the code developer.')	! This should not happen.
        ! isOptimumFound = .true.	! This is visited only once.

        write(errStr, '(a,es16.6,es16.6)') "S/N of previous/current are " &
           , prevSN, curSN
        call message(AppMsg, Verbose)
        write(errStr, '(a,i8,a,i8)') "The local maximum S/N is achieved at i=" &
           , prevLoc, " nbin=",count(prevMask)
        call message(AppMsg, Noisy)
        bestSN = prevSN

        optimumY = aryBRateCut(iDiag-1)	! == min(prevY+100*epsilon(prevY), (prevY+prevprevY)/2.0)
!write(errStr, '("prevY, optimumY, prevprevY = ",es20.12,es20.12,es20.12)') &
! prevY, optimumY, prevprevY
!call message(appMsg, 0)
!call warning('messageDump')	! Warning is raised.
        ! NOTE: The former (with a smaller coefficient, maybe 1) may be enough,
        !  or at least it is definitely enough, numerically speaking.
        !  However a simple x+epsilon(x) can be too small
        !  to distinguish in STDOUT by human eyes.
        ! On the other hand, equally, the latter would be probably enough.
        !  However if a mask is specified for this routine, then the latter
        !  can be very large, whereas the user may not want such a large
        !  value for the later use, for example, when the user uses the value
        !  as an input for SAS-task tabgtigen, where the mask that the user
        !  specified for this task may not be taken into account.

!print *,"nDup=",nDup,", count(curMask)=",count(curMask),", count(prevMask)=",count(prevMask) 
        if (present(optimumSize)) then
          optimumSize = count(prevMask)
          optimumSizeLocMax = optimumSize
        end if
        !if (present(optimumSize)) optimumSize = count(curMask) + nDup	! Plus nDup because the current removal of the index must be invalid and restored.
        if (present(mask)) then
           mask = prevMask
           locMaxMask = mask
        end if
      end if	! if ((curSN < prevSN) .and. (bestSN < prevSN)) then

        ! if (localDoDump) then
          ! Continue the loop.
      nDup = 1
      prevSN = curSN
      prevprevY = prevY 
      prevY  = curY
      prevMask = curMask
      prevLoc = curLoc
        ! else
        !   return	! Found the solution!
        ! end if
      ! end if
    end do	! do nbin=sizeMaskIn, lowerNbin, -1	! (lowerNbin >= 2) is guaranteed.

! print *, 'DEBUG: sizeMaskIn, lowerNbin =', sizeMaskIn, lowerNbin
! write(errStr, '(a,i10,i10)') 'DEBUG: sizeMaskIn, lowerNbin =', sizeMaskIn, lowerNbin
! call message(appMsg, 0)
! print *, 'DEBUG: curSN, bestSN=', curSN, bestSN
    ! if (isOptimumFound) then	! The optimum value is already found.
    ! if (prevSN < bestSN) then	! The last bin is not the optimum, hence the optimum should have been found already.  n.b., curSN may not be initialised here, hence using prevSN.
    if (present(bkgDiagn)) then
      if (allocated(bkgDiagn%aryBRawCut)) then
        call message(AppMsg, Noisy, 'bkgoptrate_aux: bkgDiagn already allocated.')	! Perhaps not intended...
      else
        call initialiseBkgoptrateDiagnosisT(iDiag, bkgDiagn)
      end if
      bkgDiagn%aryBRawCut(:)  = aryBRawCut(1:iDiag)
      bkgDiagn%aryBRateCut(:) = aryBRateCut(1:iDiag)
      bkgDiagn%aryRate(:)     = aryRate(1:iDiag)
      bkgDiagn%arySN(:)   = arySN(  1:iDiag)
      bkgDiagn%aryNBin(:) = aryNBin(1:iDiag)
    end if

    if (prevSN <= bestSN) then	! The last bin is not the optimum, hence the optimum should have been found already - it is the most normal situation.  Note for developers: curSN may not be initialised here, hence using prevSN.
! print *, 'DEBUG: return....: prevSN, bestSN =', prevSN, bestSN
      return
    end if
! print *, 'DEBUG: No maximum is found....'

    ! There is no maximum, in other words the smaller number of bins are used,
    ! better the statistics will be.
    if (nDup > 1) then
      if (lowerNbin <= 2) then	! Minimum possible lowerNbin
        if (nDup > sizeMaskIn/3) then
          ! Over one-thirds of the time-series have the same and low Y-values.
          ! Therefore it must be desirable to return the cut-off above the value.
          ! In practice, this must be an unrealistic situation, though.
          write(errStr, '(a,es14.8,a)') 'Failed to find the maximum S/N.  &
             &However, the significant number of lowest bins are found during the search &
             &to have the same counts (or rate): ', curY &
             , '.  They occupy more than one-third of the total number &
             &of bins in the input time-series, therefore this task regards &
             &the value as the cut-off threshold.'
          call warning('regardLowestBinsAsOptimum')	! Warning is raised.

          optimumY = min(prevY+100*epsilon(prevY), (prevY+prevprevY)/2.0)
          if (present(optimumSize)) optimumSize = count(prevMask)
          if (present(mask)) then
            mask = prevMask
          end if
!!!!! status = 1	!???????
          return

        else
          ! Do nothing (ie., proceed)
        end if	! if (nDup > sizeMaskIn/3) then

      else	! if (lowerNbin <= 2) then	! Minimum possible lowerNbin
        write(errStr, '(a,es14.8)') 'You may want to retry &
           &with a lower mintimeratio (in the command-line argument), &
           &because there is a possibility to find the optimum value with it &
           &given the fact more than one lowest bins are found during the search &
           &to have the same counts (or rate): ', curY
        if (present(status)) then
          call message(AppMsg, Sparse)
        else
          call warning('recommendedLowerMinTimeRatio')	! Then Error is raised.
        end if
      end if	! if (lowerNbin <= 2) then	! Minimum possible lowerNbin
    end if	! if (nDup > 1) then

    !! The fact the code reaches hereafter means that the result was erroneous in one way or another.
    if (present(status)) then
      call message(appMsg, Sparse, 'The lowest-count bin has the maximum S/N.')
      if (optimumY < 0) then	! optimumY has not been found at all, that is, no local maximum for S/N was found, i.e., S/N is monotoneously better for the smaller count (or, possibly there may be a local minimum for S/N, but not maximum).
        status = -3
      else
        status = -2
      end if

    else
      call error('noThresholdFound' &
         , 'Smaller the number of bins are used, better the statistics will'&
         //' be in the given condition.')
    end if

    if (present(mask)) mask = .false.

  end subroutine findOptimumYStrict


  !---------------------------------------------------------------------
  subroutine findOptimumYStrictClean(tsCnt, optimumY &
     , optimumSize, tsRate, aryFracexp, mask, lowerNbinRatio, status &
     , bkgDiagn, doDump)
  !---------------------------------------------------------------------
  ! *** Description ***
  !  Find the optimum cut of Y-value (Count or Rate) that gives the maximum
  ! S/N ratio for the input time-series, which should be the background
  ! time-series, where
  !    S/N =~ Tot_EXP / SQRT(Tot_Cnt)  (x Rate (=Constant) to give the counts).
  !  Then you (the calling routine) can exclude those high-background periods
  ! from another source time-series, based on the optimum background cut.
  ! 
  ! *** Arguments ***
  !  tsCnt(:) : Array of Counts for time bin.
  !    This is used to estimate the S/N.
  !    Note the unit has to be count (or alike).  Threrefore, if the original
  !    time-series is in unit of Rate, tsCnt(:) has to be calculated and
  !    given to this subroutine by the caller.
  !  optimumY   : The calculated optimum Y-value.
  !  optimumSize : The size of tsCnt(:) when it gives the best S/N (=optimumY).
  !    This is equivalent to count(mask) (after mask is modified in this routine).
  !  tsRate(:) : Array of Rate for each time bin (OPTIONAL).  If not given,
  !    (1) if aryFracexp is not given either, tsRate is set to be tsCnt,
  !    (2) if aryFracexp given, it is calculated accordingly.
  !    tsRate and aryFracexp should not be independent, but this subroutine
  !    does not check consistency when both the arguments are given.
  !    This is used to estimate the optimum threshold.
  !  aryFracexp(:) : Array of FRACEXP (OPTIONAL).  If not given,
  !    (1) if tsRate is not given either, aryFracexp(:) is set to be all 1.0,
  !    (2) if tsRate given, it is calculated accordingly.
  !    tsRate and aryFracexp should not be independent, but this subroutine
  !    does not check consistency when both the arguments are given.
  !  mask(:)    : Mask Array, indicating which indices in tsCnt(:) are valid.
  !    This mask is modified in the subroutine, and the returned mask shows
  !    true for all the valid values when the optimum value is found.
  !  lowerNbinRatio : Threshold in ratio for the number of bin used,
  !    where the calculation stops.  For example, if it is 0.05,
  !    this routine does not search for the optimum Y-value beyond the point,
  !    where the number of bins used is 5% of the entire bin number
  !    after the given mask is applied; that is, for example,
  !    when the size(tsCnt) is 1000 and the given mask has 80% of .true.,
  !    this routine does not search for the optimum value for the number
  !    of bins used being less than 40 (=1000*0.8*0.05), and if it happens,
  !    either error('noThresholdFound') is raised or status=-2 is returned.
  !    The default is 0.0.
  !  status     : Result (fitting) status indicator (OPTIONAL).
  !    0 for normal end,
  !    -2 if the optimum values is not found even when the search reaches the end
  !    (in that case, if status is not given, Error is raised.).
  !  doDump     : For debugging - it dumps all the S/Ns if True.
  !    Note the returned values are most likely to be incorrect if True.
  ! 
  ! *** ALGORITHM ***
  !   0. Prepare TSeries(:), TRate(:) and Mask(:).
  !   1. Calculate the S/N in TSeries, based on Mask.
  !   2. Find the maximum value and its locations, namely the indices, in TRate(:).
  !   3. Set Mask(:) false for the index(ices), which have the maximum value,
  !  and calculate the S/N in TSeries, based on (revised) Mask.
  !   4. If the new S/N is larger than the previous S/N,
  !  then repeat the processing from (2).  If smaller,
  !  then the previous maximum value and Mask are the optimum ones (best S/N),
  !  so return them.
  !
  ! *** TODO ***
  !  * At the time of writing, the minimization (maximization) algorithm
  ! in this subroutine is extremely crude and inefficient --- should
  ! be rewritten with the use of a proper minimization routine, ideally.
  !
  !  * It would be simpler if all the input time-series are sorted
  ! in order of Count (or Rate or corrected Rate with FRACEXP,
  ! depending on the requirement) before they are processed.
  ! Perhaps use ssclib/sort_mod for it.
  !

    implicit none

    real(single),   intent(in) :: tsCnt(:)
    real(single),   intent(out) :: optimumY
    integer(int32), intent(out), optional :: optimumSize
    real(single),   intent(in),  optional :: tsRate(size(tsCnt)), aryFracexp(size(tsCnt))
    ! logical,        intent(in),  optional :: isRate	! This has to be provided if aryExp(:) is given.
    logical,        intent(inout), optional :: mask(size(tsCnt))
    ! logical,        intent(out), optional :: maskOut(size(tsCnt))
    real(single),   intent(in),  optional :: lowerNbinRatio
    integer,        intent(out), optional :: status
    type(BkgoptrateDiagnosisT), intent(out), optional :: bkgDiagn
    logical,        intent(in),  optional :: doDump

    logical, dimension(size(tsCnt)) :: prevMask, curMask
    logical :: isFracexpUsed, localDoDump	! , isOptimumFound
    real(single), dimension(size(tsCnt)) :: localTsRate, aryExp, aryBRawCut, aryBRateCut, arySN, aryY
    integer,      dimension(size(tsCnt)) :: aryNBin
    real(single) :: prevSN, curSN, bestSN, prevprevY, prevY, curY, nextY, tmpY, bestY, bestPrevY
    integer(int32) :: nbin, sizeMaskIn, lowerNbin, nDup, iDiag &
       , iBestDiag !, iMax
    integer :: arMaxloc(1), prevLoc, curLoc

logical, parameter :: localIsTest = .false.

    !-------------------------------------------------------------------
    ! Start

    ! *** Algorithm (see also the above comment for the overview) ***
    !  Assumption: The function to give S/N is monotonic, namely having
    ! only one maximum value.
    !
    !  (1) Set False at the X (prevLoc) for the maximum Y (Rate) in the current mask (curMask).
    !  (2) Find the index X for the maximum Y (Rate) with the new mask (curLoc),
    !    as well as the maximum Y-value (curYRate).
    !  (3a) If curY==prevY then repeat from (1) after resetting prevLoc as curLoc.
    !  (3b) If not, proceed below.
    !  (4) Calculate S/N (curSN) for TS with the new mask (curMask).
    !  (5a) If (curSN > prevSN) then repeat from (1) after resetting prev*.
    !  (5b) If not, then returns prev* (nb. the returned Y-value is slightly larger than prevYRate)
    !  (6) If fails to find the maximum for SN, then it returns status=-2 or raises an error.
    !    However, even in that case, if over one-third of the lowest bins
    !    have the same Y(Rate)-value, AND if the lowerNbinRatio given is the minimum,
    !    then the Y(Rate)-value (or strictly speaking a value slightly higher
    !    than that) is returned and this subroutines ends normally.
    !    This case may happen in, such as, uniform time-series or step-function-like ones.
    !
    ! *** Notation ***
    !  (prev|cur)Mask : The mask file used to calculate the values.
    !  (prev|cur)SN   : Previous/Current SN when the above mask is applied.
    !  (prev|cur)Y    : Maximum Y(Rate)-value (when the above mask is applied).
    !  (prev|cur)Loc  : Its index             (when the above mask is applied).
    !  prevprevY      : Previous previous maximum Y(Rate)-value.

    ! isOptimumFound = .false.

    optimumY = -1
    iBestDiag    = 0
    bestSN    = -9999.0
    localTsRate(:) = tsCnt
    aryExp(:) = 1.0
    isFracexpUsed = .false.
    if (present(status)) status = 0
    if (present(optimumSize)) optimumSize = -1
    if (present(tsRate)) then
      isFracexpUsed = .true.
      localTsRate = tsRate
      if (present(aryFracexp)) then
        aryExp(:) = aryFracexp
      else
        where(localTsRate .ne. 0) aryExp(:) = tsCnt/localTsRate
        ! It is impossible to guess aryExp() when tsRate (tsCnt) is zero.
      end if
    else if (present(aryFracexp)) then
      isFracexpUsed = .true.
        aryExp(:) = aryFracexp
        where(aryExp .ne. 0) localTsRate(:) = tsCnt/aryExp
        ! It makes no sense to determine tsRate when aryExp is zero.
    end if
    ! if (present(maskOut)) maskOut = .false.

    localDoDump = .false.
    if (present(bkgDiagn)) then
      localDoDump = .true.
    else if (present(doDump)) then
      localDoDump = doDump
    end if

    curMask = .true.
    if (present(mask)) curMask = mask
    sizeMaskIn = count(curMask)
    prevMask = curMask
    if (isFracexpUsed) then
      ! prevSN = getSN2ofTs(tsCnt, aryExp, .false., prevMask)
      prevSN = getSN2ofTs(tsCnt, prevMask, aryExp)
    else
      prevSN = getSN2ofTs(tsCnt, prevMask)
      ! prevSN = sizeMaskIn / sqrt(sum(tsCnt, MASK=prevMask))	! S/N calculation.
    end if

    if (present(lowerNbinRatio)) then
      lowerNbin = nint(sizeMaskIn * lowerNbinRatio)
      if ((lowerNbin == 0) .or. (lowerNbin == 1)) then
        lowerNbin = 2	! when lowerNbinRatio=0.0
      else if (lowerNbin < 0) then
        ! lowerNbinRatio has to be zero or positive.
        call error('Bug_lowerNbinRatio', 'Contact the code developer.')
      end if
    else
      lowerNbin = 2
    end if

    if (verbosity(AppMsg) >= Noisy) then
      if (present(lowerNbinRatio)) then
        write(errStr, '(a,i8,a,i8,a,i6,a,f7.4,a)') "Start findOptimumY(): Size TS(" &
           , size(tsCnt), "), mask(", sizeMaskIn &
           , "), lowerNbin=", lowerNbin, " (lowerNbinRatio=", lowerNbinRatio &
           , "  which will be applied further)."
      else
        write(errStr, '(a,i8,a,i8,a,i6,a)') "Start findOptimumY():  Size TS(" &
           , size(tsCnt), "), mask(", sizeMaskIn &
           , "), lowerNbin=", lowerNbin, " (lowerNbinRatio=UNSPECIFIED)."
      end if
      call message(AppMsg, Noisy)
    end if

    if (localIsTest) then
      print *,"DEBUG: nbin=",count(prevMask),"curSN2=",prevSN,"in Start."
      ! print *,"DEBUG: nbin=",count(curMask),"curSN2=",prevSN,"in Start."
    end if

    arMaxloc = maxloc(localTsRate, prevMask)
    prevLoc = arMaxloc(1)
    prevY = localTsRate(prevLoc)	! == maxval(localTsRate, prevMask)
    ! prevY = prevY + abs(prevY)/100.0	! slightly larger than the first maximum value.
    ! prevY = prevY + abs(epsilon(prevY)) * 4	! slightly larger than the first maximum value.
    prevprevY = prevY 
    bestY     = prevY
    bestPrevY = prevY
    nDup = 1

    aryBRawCut(:)  = -1.0	! Initial value.
    aryBRateCut(:) = -1.0
    arySN(:)   = -1.0
    aryNBin(:) = -1
    aryY(:)    = -1.0
    iDiag = 1
    aryBRawCut(iDiag) = tsCnt(prevLoc) + 100*epsilon(tsCnt(prevLoc))	! In count.
    aryBRateCut(iDiag) = prevY + 100*epsilon(prevY)	! 100 is an arbitrary number.
    aryY(iDiag)    = prevY
    arySN(iDiag)   = prevSN
    aryNBin(iDiag) = count(prevMask)

    do nbin=sizeMaskIn, lowerNbin, -1	! (lowerNbin >= 2) is guaranteed.
      ! nbin is the number of bins to be used in the estimate in each cycle.

      curMask(prevLoc) = .false.
      arMaxloc = maxloc(localTsRate, curMask)	! To calculate curLoc.
      curLoc = arMaxloc(1)
      curY = localTsRate(curLoc)	! Maximum Y for curMask

      if (prevY == curY) then	! The current maximum Y-value is identical to the previous Y-value, so should be taken away before the next evaluation.
        prevLoc = curLoc 
        nDup = nDup + 1
        if (localIsTest) then
          print *,"DEBUG: nbin=",count(curMask),"                    Removed-(i,Y)=",prevLoc,prevY
        end if
        cycle
      end if

      if (isFracexpUsed) then
        !curSN = getSN2ofTs(tsCnt, aryExp, .false., curMask)
        curSN = getSN2ofTs(tsCnt, curMask, aryExp)
        ! NOTE: This formula can replace the one in else-clause, in other words,
        !  this IF clause itself is not essential, however it would be slower.
      else
        curSN = getSN2ofTs(tsCnt, curMask)
      end if

      iDiag = iDiag + 1
      if (aryBRawCut(iDiag) > 0) call error('bugAryBcut','Contact the code developer.')	! This should not happen.
      ! aryBRawCut(iDiag) = (curY + aryBRawCut(iDiag-1))/2.0
      aryBRawCut(iDiag) = tsCnt(curLoc) + 2*epsilon(tsCnt(curLoc))
      aryBRateCut(iDiag) = min(curY+100*epsilon(curY), (curY+prevY)/2.0)	! The threshold Y must be larger than the Y that gives the maximum S/N (say, by 100 times of epsilon()), yet be smaller than the next possible Y that gives a different (and by definition smaller) S/N.
      aryY(iDiag)    = curY
      arySN(iDiag)   = curSN
      aryNBin(iDiag) = count(curMask)
      if (curSN > prevSN) then
        iBestDiag = iDiag
        bestSN = curSN
        bestY     = curY	! Y value that gives the best S/N
        bestPrevY = prevY	! Y value that is the previous of bestY, namely, that is larger than bestY, yet smaller than any other larger Ys.
        if (present(optimumSize)) optimumSize = aryNBin(iDiag)	! == count(curMask)
        if (present(mask)) then
           mask = curMask	! Because of this "mask", this "if (curSN > prevSN)" clause is necessary...  But this is a little time-consuming, hence should ideally be improved.
        end if
      end if

      !THIS NOTE IS INCORRECT!!!!!
      ! NOTE: Here, count(curMask) should be equal to (nbin-1).
      !  Therefore the following equation should do the same job,
      !  and would be faster due to no overhead.
      ! curSN = (nbin-1) / sqrt(sum(tsCnt, MASK=curMask))
      !
      !  But currently I use the above function just to be conservative and safe.

      if (localIsTest) then
        print *,"DEBUG: nbin=",count(curMask),"curSN2=",curSN,"Removed-(i,Y)=",prevLoc,prevY
      end if

      prevSN  = curSN
      prevY   = curY
      prevLoc = curLoc

      ! if ((isOptimumFound) .or. (curSN > prevSN)) then
      !   nDup = 1
      !   prevSN = curSN
      !   prevprevY = prevY 
      !   prevY  = curY
      !   prevMask = curMask
      !   prevLoc = curLoc
      !   cycle	! At the time of writing, this is redundant, but just to play safe.
      ! 
      ! else
      !   ! The previous attempt gives the better S/N,
      !   ! so it is regarded as the best S/N value and is returned.
      ! 
      !   if (isOptimumFound) call error('bugOptimum','Contact the code developer.')	! This should not happen.
      !   isOptimumFound = .true.	! This is visited only once.
      ! 
      !   write(errStr, '(a,es16.6,es16.6)') "S/N of previous/current are " &
      !      , prevSN, curSN
      !   call message(AppMsg, Verbose)
      !   write(errStr, '(a,i8,a,i8)') "The best S/N is achieved at i=" &
      !      , prevLoc, " nbin=",count(prevMask)
      !      !, curLoc, " nbin=",count(curMask)+nDup
      !   call message(AppMsg, Noisy)
      ! 
      !   optimumY = min(prevY+100*epsilon(prevY), (prevY+prevprevY)/2.0)
      !   ! NOTE: The former (with a smaller coefficient, maybe 1) may be enough,
      !   !  or at least it is definitely enough, numerically speaking.
      !   !  However a simple x+epsilon(x) can be too small
      !   !  to distinguish in STDOUT by human eyes.
      !   ! On the other hand, equally, the latter would be probably enough.
      !   !  However if a mask is specified for this routine, then the latter
      !   !  can be very large, whereas the user may not want such a large
      !   !  value for the later use, for example, when the user uses the value
      !   !  as an input for SAS-task tabgtigen, where the mask that the user
      !   !  specified for this task may not be taken into account.
      ! 
      ! !print *,"nDup=",nDup,", count(curMask)=",count(curMask),", count(prevMask)=",count(prevMask) 
      !   if (present(optimumSize)) optimumSize = count(prevMask)
      !   !if (present(optimumSize)) optimumSize = count(curMask) + nDup	! Plus nDup because the current removal of the index must be invalid and restored.
      !   if (present(mask)) then
      !      mask = prevMask
      !      ! maskOut = prevMask
      !   end if
      ! 
      !   if (localDoDump) then
      !     ! Continue the loop.
      !     nDup = 1
      !     prevSN = curSN
      !     prevprevY = prevY 
      !     prevY  = curY
      !     prevMask = curMask
      !     prevLoc = curLoc
      !   else
      !     return	! Found the solution!
      !   end if
      ! end if
    end do

    ! arMaxloc = maxloc(arySN(1:iDiag))
    ! iMax = arMaxloc(1)
    ! tmpY = arySN(iMax)
    ! if (iDiag <= 1) then
    !   prevY = tmpY 
    !   nextY = tmpY 
    ! else
    !   if (iMax <= 1) then
    !     prevY = tmpY 
    !     nextY = arySN(2)
    !   else if (iMax >= iDiag) then
    !     prevY = arySN(iDiag - 1)
    !     nextY = tmpY 
    !   else
    !     prevY = arySN(iDiag - 1)
    !     nextY = arySN(iDiag + 1)
    !   end if
    ! end if

    if (iBestDiag < iDiag) then
    ! if (isOptimumFound) then	! The optimum value is already found.

      optimumY = min(bestY+100*epsilon(bestY), (bestY+bestPrevY)/2.0)
      if (present(optimumSize)) then
        ! NOTE: optimumSize and mask should be already defined appropriately, except for some special cases.
        if (optimumSize < 0) then
          optimumSize = aryNBin(1)	! == count(curMask)
          ! optimumSize = aryNBin(iBestDiag)	! == count(curMask)
          call warning('maybeUniformTS' &
             , 'optimumSize must be redefined here, perhaps due to the uniform input TS.')
        end if
      end if

      if (present(bkgDiagn)) then
        if (allocated(bkgDiagn%aryBRawCut)) then
          call message(AppMsg, Noisy, 'bkgoptrate_aux: bkgDiagn already allocated.')	! Perhaps not intended...
        else
          call initialiseBkgoptrateDiagnosisT(iDiag, bkgDiagn)
        end if
        bkgDiagn%aryBRawCut(:)  = aryBRawCut(1:iDiag)
        bkgDiagn%aryBRateCut(:) = aryBRateCut(1:iDiag)
        bkgDiagn%arySN(:)   = arySN(  1:iDiag)
        bkgDiagn%aryNBin(:) = aryNBin(1:iDiag)
      end if

      return
    end if

    ! There is no maximum, in other words the smaller number of bins are used,
    ! better the statistics will be.
    if (nDup > 1) then
      if (lowerNbin <= 2) then	! Minimum possible lowerNbin
        if (nDup > sizeMaskIn/3) then
          ! Over one-thirds of the time-series have the same and low Y-values.
          ! Therefore it must be desirable to return the cut-off above the value.
          ! In practice, this must be an unrealistic situation, though.
          write(errStr, '(a,es14.8,a)') 'Failed to find the maximum S/N.  &
             &However, the significant number of lowest bins are found during the search &
             &to have the same counts (or rate): ', curY &
             , '.  They occupy more than one-third of the total number &
             &of bins in the input time-series, therefore this task regards &
             &the value as the cut-off threshold.'
          call warning('regardLowestBinsAsOptimum')

          ! optimumY = min(prevY+100*epsilon(prevY), (prevY+prevprevY)/2.0)
          optimumY = min(bestY+100*epsilon(bestY), (bestY+bestPrevY)/2.0)
          if (present(optimumSize)) optimumSize = count(prevMask)
          print *,'nbin, lowerNbin = ', nbin, lowerNbin
          print *,'c(prev,cur) = ', count(prevMask), count(curMask)
          if (present(mask)) then
            mask = prevMask
            ! maskOut = prevMask
          end if
          return

        else
          ! Do nothing (ie., proceed).  Error will be raised immediately afterward.
        end if

      else	! if (lowerNbin <= 2) then
        write(errStr, '(a,es14.8)') 'You may want to retry &
           &with a lower mintimeratio (in the command-line argument), &
           &because there is a possibility to find the optimum value with it &
           &given the fact more than one lowest bins are found during the search &
           &to have the same counts (or rate): ', curY
        if (present(status)) then
          call message(AppMsg, Sparse)
        else
          call warning('recommendedLowerMinTimeRatio')	! Then Error will be raised immediately afterward.
        end if
      end if	! if (lowerNbin <= 2) then ... else ...
    end if

    if (present(status)) then
      call message(AppMsg, Sparse, 'Failed to find the optimum threshold.')
      status = -2
    else
      call error('noThresholdFound' &
         , 'The smaller number of bins are used, better the statistics will'&
         //' be in the given condition.')
    end if

    if (present(mask)) mask = .false.

  end subroutine findOptimumYStrictClean


  !---------------------------------------------------------------------
  subroutine findOptimumYCrude(tsCnt, optimumY &
     , optimumSize, tsRate, aryFracexp, mask, lowerNbinRatio, status &
     , bkgDiagn, doDump)
  !---------------------------------------------------------------------
  ! *** Description ***
  !  Find the optimum cut of Y-value (Count or Rate) that gives the maximum
  ! S/N ratio for the input time-series, which should be the background
  ! time-series, where
  !    S/N =~ Tot_EXP / SQRT(Tot_Cnt)  (x Rate (=Constant) to give the counts).
  !  Then you (the calling routine) can exclude those high-background periods
  ! from another source time-series, based on the optimum background cut.
  ! 
  ! *** Arguments ***
  !  tsCnt(:) : Array of Counts for time bin.
  !    This is used to estimate the S/N.
  !    Note the unit has to be count (or alike).  Threrefore, if the original
  !    time-series is in unit of Rate, tsCnt(:) has to be calculated and
  !    given to this subroutine by the caller.
  !  optimumY   : The calculated optimum Y-value.
  !  optimumSize : The size of tsCnt(:) when it gives the best S/N (=optimumY).
  !    This is equivalent to count(mask) (after mask is modified in this routine).
  !  tsRate(:) : Array of Rate for each time bin (OPTIONAL).  If not given,
  !    (1) if aryFracexp is not given either, tsRate is set to be tsCnt,
  !    (2) if aryFracexp given, it is calculated accordingly.
  !    tsRate and aryFracexp should not be independent, but this subroutine
  !    does not check consistency when both the arguments are given.
  !    This is used to estimate the optimum threshold.
  !  aryFracexp(:) : Array of FRACEXP (OPTIONAL).  If not given,
  !    (1) if tsRate is not given either, aryFracexp(:) is set to be all 1.0,
  !    (2) if tsRate given, it is calculated accordingly.
  !    tsRate and aryFracexp should not be independent, but this subroutine
  !    does not check consistency when both the arguments are given.
  !  mask(:)    : Mask Array, indicating which indices in tsCnt(:) are valid.
  !    This mask is modified in the subroutine, and the returned mask shows
  !    true for all the valid values when the optimum value is found.
  !  lowerNbinRatio : Threshold in ratio for the number of bin used,
  !    where the calculation stops.  For example, if it is 0.05,
  !    this routine does not search for the optimum Y-value beyond the point,
  !    where the number of bins used is 5% of the entire bin number
  !    after the given mask is applied; that is, for example,
  !    when the size(tsCnt) is 1000 and the given mask has 80% of .true.,
  !    this routine does not search for the optimum value for the number
  !    of bins used being less than 40 (=1000*0.8*0.05), and if it happens,
  !    either error('noThresholdFound') is raised or status=-2 is returned.
  !    The default is 0.0.
  !  status     : Result (fitting) status indicator (OPTIONAL).
  !    0 for normal end,
  !    -2 if the optimum values is not found even when the search reaches the end
  !    (in that case, if status is not given, Error is raised.).
  !  doDump     : For debugging - it dumps all the S/Ns if True.
  !    Note the returned values are most likely to be incorrect if True.
  ! 
  ! *** ALGORITHM ***
  !   0. Prepare TSeries(:), TRate(:) and Mask(:).
  !   1. Calculate the S/N in TSeries, based on Mask.
  !   2. Find the maximum value and its locations, namely the indices, in TRate(:).
  !   3. Set Mask(:) false for the index(ices), which have the maximum value,
  !  and calculate the S/N in TSeries, based on (revised) Mask.
  !   4. If the new S/N is larger than the previous S/N,
  !  then repeat the processing from (2).  If smaller,
  !  then the previous maximum value and Mask are the optimum ones (best S/N),
  !  so return them.
  !
  ! *** TODO ***
  !  * At the time of writing, the minimization (maximization) algorithm
  ! in this subroutine is extremely crude and inefficient --- should
  ! be rewritten with the use of a proper minimization routine, ideally.
  !
  !  * It would be simpler if all the input time-series are sorted
  ! in order of Count (or Rate or corrected Rate with FRACEXP,
  ! depending on the requirement) before they are processed.
  ! Perhaps use ssclib/sort_mod for it.
  !

    implicit none

    real(single),   intent(in) :: tsCnt(:)
    real(single),   intent(out) :: optimumY
    integer(int32), intent(out), optional :: optimumSize
    real(single),   intent(in),  optional :: tsRate(size(tsCnt)), aryFracexp(size(tsCnt))
    ! logical,        intent(in),  optional :: isRate	! This has to be provided if aryExp(:) is given.
    logical,        intent(inout), optional :: mask(size(tsCnt))
    ! logical,        intent(out), optional :: maskOut(size(tsCnt))
    real(single),   intent(in),  optional :: lowerNbinRatio
    integer,        intent(out), optional :: status
    type(BkgoptrateDiagnosisT), intent(out), optional :: bkgDiagn
    logical,        intent(in),  optional :: doDump

    logical, dimension(size(tsCnt)) :: prevMask, curMask
    logical :: isFracexpUsed, localDoDump, isOptimumFound
    real(single), dimension(size(tsCnt)) :: localTsRate, aryExp, aryBRawCut, aryBRateCut, arySN
    integer,      dimension(size(tsCnt)) :: aryNBin
    real(single) :: prevSN, curSN, prevprevY, prevY, curY
    integer(int32) :: nbin, sizeMaskIn, lowerNbin, nDup, iDiag
    integer :: arMaxloc(1), prevLoc, curLoc

logical, parameter :: localIsTest = .false.

    !-------------------------------------------------------------------
    ! Start

    ! *** Algorithm (see also the above comment for the overview) ***
    !  Assumption: The function to give S/N is monotonic, namely having
    ! only one maximum value.
    !
    !  (1) Set False at the X (prevLoc) for the maximum Y (Rate) in the current mask (curMask).
    !  (2) Find the index X for the maximum Y (Rate) with the new mask (curLoc),
    !    as well as the maximum Y-value (curYRate).
    !  (3a) If curY==prevY then repeat from (1) after resetting prevLoc as curLoc.
    !  (3b) If not, proceed below.
    !  (4) Calculate S/N (curSN) for TS with the new mask (curMask).
    !  (5a) If (curSN > prevSN) then repeat from (1) after resetting prev*.
    !  (5b) If not, then returns prev* (nb. the returned Y-value is slightly larger than prevYRate)
    !  (6) If fails to find the maximum for SN, then it returns status=-2 or raises an error.
    !    However, even in that case, if over one-third of the lowest bins
    !    have the same Y(Rate)-value, AND if the lowerNbinRatio given is the minimum,
    !    then the Y(Rate)-value (or strictly speaking a value slightly higher
    !    than that) is returned and this subroutines ends normally.
    !    This case may happen in, such as, uniform time-series or step-function-like ones.
    !
    ! *** Notation ***
    !  (prev|cur)Mask : The mask file used to calculate the values.
    !  (prev|cur)SN   : Previous/Current SN when the above mask is applied.
    !  (prev|cur)Y    : Maximum Y(Rate)-value (when the above mask is applied).
    !  (prev|cur)Loc  : Its index             (when the above mask is applied).
    !  prevprevY      : Previous previous maximum Y(Rate)-value.

    isOptimumFound = .false.

    optimumY = -1
    localTsRate(:) = tsCnt
    aryExp(:) = 1.0
    isFracexpUsed = .false.
    if (present(status)) status = 0
    if (present(optimumSize)) optimumSize = -1
    if (present(tsRate)) then
      isFracexpUsed = .true.
      localTsRate = tsRate
      if (present(aryFracexp)) then
        aryExp(:) = aryFracexp
      else
        where(localTsRate .ne. 0) aryExp(:) = tsCnt/localTsRate
        ! It is impossible to guess aryExp() when tsRate (tsCnt) is zero.
      end if
    else if (present(aryFracexp)) then
      isFracexpUsed = .true.
        aryExp(:) = aryFracexp
        where(aryExp .ne. 0) localTsRate(:) = tsCnt/aryExp
        ! It makes no sense to determine tsRate when aryExp is zero.
    end if
    ! if (present(maskOut)) maskOut = .false.

    localDoDump = .false.
    if (present(bkgDiagn)) then
      localDoDump = .true.
    else if (present(doDump)) then
      localDoDump = doDump
    end if

    curMask = .true.
    if (present(mask)) curMask = mask
    sizeMaskIn = count(curMask)
    prevMask = curMask
    if (isFracexpUsed) then
      ! prevSN = getSN2ofTs(tsCnt, aryExp, .false., prevMask)
      prevSN = getSN2ofTs(tsCnt, prevMask, aryExp)
    else
      prevSN = getSN2ofTs(tsCnt, prevMask)
      ! prevSN = sizeMaskIn / sqrt(sum(tsCnt, MASK=prevMask))	! S/N calculation.
    end if

    if (present(lowerNbinRatio)) then
      lowerNbin = nint(sizeMaskIn * lowerNbinRatio)
      if ((lowerNbin == 0) .or. (lowerNbin == 1)) then
        lowerNbin = 2	! when lowerNbinRatio=0.0
      else if (lowerNbin < 0) then
        ! lowerNbinRatio has to be zero or positive.
        call error('Bug_lowerNbinRatio', 'Contact the code developer.')
      end if
    else
      lowerNbin = 2
    end if

    if (verbosity(AppMsg) >= Noisy) then
      if (present(lowerNbinRatio)) then
        write(errStr, '(a,i8,a,i8,a,i6,a,f7.4,a)') "Start findOptimumY(): Size TS(" &
           , size(tsCnt), "), mask(", sizeMaskIn &
           , "), lowerNbin=", lowerNbin, " (lowerNbinRatio=", lowerNbinRatio &
           , "  which will be applied further)."
      else
        write(errStr, '(a,i8,a,i8,a,i6,a)') "Start findOptimumY():  Size TS(" &
           , size(tsCnt), "), mask(", sizeMaskIn &
           , "), lowerNbin=", lowerNbin, " (lowerNbinRatio=UNSPECIFIED)."
      end if
      call message(AppMsg, Noisy)
    end if

    if (localIsTest) then
      print *,"DEBUG: nbin=",count(prevMask),"curSN2=",prevSN,"in Start."
      ! print *,"DEBUG: nbin=",count(curMask),"curSN2=",prevSN,"in Start."
    end if

    arMaxloc = maxloc(localTsRate, prevMask)
    prevLoc = arMaxloc(1)
    prevY = localTsRate(prevLoc)	! == maxval(localTsRate, prevMask)
    ! prevY = prevY + abs(prevY)/100.0	! slightly larger than the first maximum value.
    ! prevY = prevY + abs(epsilon(prevY)) * 4	! slightly larger than the first maximum value.
    prevprevY = prevY 
    nDup = 1

    aryBRawCut(:)  = -1.0	! Initial value.
    aryBRateCut(:) = -1.0
    arySN(:)   = -1.0
    aryNBin(:) = -1
    iDiag = 1
    aryBRawCut(iDiag) = tsCnt(prevLoc) + 100*epsilon(tsCnt(prevLoc))	! In count.
    aryBRateCut(iDiag) = prevY + 100*epsilon(prevY)	! 100 is an arbitrary number.
    arySN(iDiag)   = prevSN
    aryNBin(iDiag) = count(prevMask)

    do nbin=sizeMaskIn, lowerNbin, -1	! (lowerNbin >= 2) is guaranteed.
      ! nbin is the number of bins to be used in the estimate in each cycle.

      curMask(prevLoc) = .false.
      arMaxloc = maxloc(localTsRate, curMask)	! To calculate curLoc.
      curLoc = arMaxloc(1)
      curY = localTsRate(curLoc)	! Maximum Y for curMask

      if (prevY == curY) then	! The current maximum Y-value is identical to the previous Y-value, so should be taken away before the next evaluation.
        prevLoc = curLoc 
        nDup = nDup + 1
        if (localIsTest) then
          print *,"DEBUG: nbin=",count(curMask),"                    Removed-(i,Y)=",prevLoc,prevY
        end if
        cycle
      end if

      if (isFracexpUsed) then
        !curSN = getSN2ofTs(tsCnt, aryExp, .false., curMask)
        curSN = getSN2ofTs(tsCnt, curMask, aryExp)
        ! NOTE: This formula can replace the one in else-clause, in other words,
        !  this IF clause itself is not essential, however it would be slower.
      else
        curSN = getSN2ofTs(tsCnt, curMask)
      end if

      iDiag = iDiag + 1
      if (aryBRawCut(iDiag) > 0) call error('bugAryBcut','Contact the code developer.')	! This should not happen.
      ! aryBRawCut(iDiag) = (curY + aryBRawCut(iDiag-1))/2.0
      aryBRawCut(iDiag) = tsCnt(curLoc) + 2*epsilon(tsCnt(curLoc))
      aryBRateCut(iDiag) = min(curY+100*epsilon(curY), (curY+prevY)/2.0)
      arySN(iDiag)   = curSN
      aryNBin(iDiag) = count(curMask)

      !THIS NOTE IS INCORRECT!!!!!
      ! NOTE: Here, count(curMask) should be equal to (nbin-1).
      !  Therefore the following equation should do the same job,
      !  and would be faster due to no overhead.
      ! curSN = (nbin-1) / sqrt(sum(tsCnt, MASK=curMask))
      !
      !  But currently I use the above function just to be conservative and safe.

      if (localIsTest) then
        print *,"DEBUG: nbin=",count(curMask),"curSN2=",curSN,"Removed-(i,Y)=",prevLoc,prevY
      end if

      if ((isOptimumFound) .or. (curSN > prevSN)) then
        nDup = 1
        prevSN = curSN
        prevprevY = prevY 
        prevY  = curY
        prevMask = curMask
        prevLoc = curLoc
        cycle	! At the time of writing, this is redundant, but just to play safe.

      else
        ! The previous attempt gives the better S/N,
        ! so it is regarded as the best S/N value and is returned.

        if (isOptimumFound) call error('bugOptimum','Contact the code developer.')	! This should not happen.
        isOptimumFound = .true.	! This is visited only once.

        write(errStr, '(a,es16.6,es16.6)') "S/N of previous/current are " &
           , prevSN, curSN
        call message(AppMsg, Verbose)
        write(errStr, '(a,i8,a,i8)') "The best S/N is achieved at i=" &
           , prevLoc, " nbin=",count(prevMask)
           !, curLoc, " nbin=",count(curMask)+nDup
        call message(AppMsg, Noisy)

        optimumY = min(prevY+100*epsilon(prevY), (prevY+prevprevY)/2.0)
        ! NOTE: The former (with a smaller coefficient, maybe 1) may be enough,
        !  or at least it is definitely enough, numerically speaking.
        !  However a simple x+epsilon(x) can be too small
        !  to distinguish in STDOUT by human eyes.
        ! On the other hand, equally, the latter would be probably enough.
        !  However if a mask is specified for this routine, then the latter
        !  can be very large, whereas the user may not want such a large
        !  value for the later use, for example, when the user uses the value
        !  as an input for SAS-task tabgtigen, where the mask that the user
        !  specified for this task may not be taken into account.

!print *,"nDup=",nDup,", count(curMask)=",count(curMask),", count(prevMask)=",count(prevMask) 
        if (present(optimumSize)) optimumSize = count(prevMask)
        !if (present(optimumSize)) optimumSize = count(curMask) + nDup	! Plus nDup because the current removal of the index must be invalid and restored.
        if (present(mask)) then
           mask = prevMask
           ! maskOut = prevMask
        end if

        if (localDoDump) then
          ! Continue the loop.
          nDup = 1
          prevSN = curSN
          prevprevY = prevY 
          prevY  = curY
          prevMask = curMask
          prevLoc = curLoc
        else
          return	! Found the solution!
        end if
      end if
    end do

    if (isOptimumFound) then	! The optimum value is already found.
      if (present(bkgDiagn)) then
        if (allocated(bkgDiagn%aryBRawCut)) then
          call message(AppMsg, Noisy, 'bkgoptrate_aux: bkgDiagn already allocated.')	! Perhaps not intended...
        else
          call initialiseBkgoptrateDiagnosisT(iDiag, bkgDiagn)
        end if
        bkgDiagn%aryBRawCut(:)  = aryBRawCut(1:iDiag)
        bkgDiagn%aryBRateCut(:) = aryBRateCut(1:iDiag)
        bkgDiagn%arySN(:)   = arySN(  1:iDiag)
        bkgDiagn%aryNBin(:) = aryNBin(1:iDiag)
      end if
      return
    end if

    ! There is no maximum, in other words the smaller number of bins are used,
    ! better the statistics will be.
    if (nDup > 1) then
      if (lowerNbin <= 2) then	! Minimum possible lowerNbin
        if (nDup > sizeMaskIn/3) then
          ! Over one-thirds of the time-series have the same and low Y-values.
          ! Therefore it must be desirable to return the cut-off above the value.
          ! In practice, this must be an unrealistic situation, though.
          write(errStr, '(a,es14.8,a)') 'Failed to find the maximum S/N.  &
             &However, the significant number of lowest bins are found during the search &
             &to have the same counts (or rate): ', curY &
             , '.  They occupy more than one-third of the total number &
             &of bins in the input time-series, therefore this task regards &
             &the value as the cut-off threshold.'
          call warning('regardLowestBinsAsOptimum')	! Then Error is raised.

          optimumY = min(prevY+100*epsilon(prevY), (prevY+prevprevY)/2.0)
          if (present(optimumSize)) optimumSize = count(prevMask)
          if (present(mask)) then
            mask = prevMask
            ! maskOut = prevMask
          end if
          return

        else
          ! Do nothing (ie., proceed)
        end if

      else
        write(errStr, '(a,es14.8)') 'You may want to retry &
           &with a lower mintimeratio (in the command-line argument), &
           &because there is a possibility to find the optimum value with it &
           &given the fact more than one lowest bins are found during the search &
           &to have the same counts (or rate): ', curY
        if (present(status)) then
          call message(AppMsg, Sparse)
        else
          call warning('recommendedLowerMinTimeRatio')	! Then Error is raised.
        end if
      end if
    end if

    if (present(status)) then
      call message(AppMsg, Sparse, 'Failed to find the optimum threshold.')
      status = -2
    else
      call error('noThresholdFound' &
         , 'The smaller number of bins are used, better the statistics will'&
         //' be in the given condition.')
    end if

    if (present(mask)) mask = .false.

  end subroutine findOptimumYCrude


  !---------------------------------------------------------------------
  subroutine calcThresholdFromAny(aryX, aryY, optimumRate &
     , optimumSize, maskSize, fracexp, timeStartEnd &
     , lowerCutoffY, lowerNbinRatio, status, bkgDiagn, doDump)
  !---------------------------------------------------------------------
  ! Description:
  !  Wrapper for findOptimumY().
  !  This subroutine adjusts aryY according to fracexp (if that is the case),
  ! makes a mask file, which is false where aryY()<=0 
  ! or where fracexp threshold (if specified so) is not satisfied,
  ! or aryX is out of range in timeStartEnd (if specified so),
  ! and finally passes the variables to findOptimumY().
  ! 
  ! NOTE of caution: If fracexp is given and its %doUse==.true. and
  !  %style="calc", then the returned optimumRate is adjusted accordingly.
  !  For example, if the FRACEXP for every bin is 0.5 and
  ! if the aryY(:) is all 10, then the returned optimumRate will be 20.
  ! In practice, most of FRACEXP is likely to be 1 or very close to 1,
  ! therefore it ie expected this should not cause any practical trouble.
  ! If it ever does, it is up to the user how to use (readjust?) the result.
  !
  ! Arguments:
  !  aryX(:) : TIME column (has to be in order of time.  eg., MJD.  Each time bin must have the same duration.).
  !  aryY(:) : RATE/COUNTS column
  !  optimumRate : The returned Rate value, which is what you want.
  !  optimumSize : The size of the array when the optimum is found.
  !  maskSize    : The size of the mask array before the search.
  !  timeStartEnd, infoFracexp : (OPTIONAL) Constraints if given.
  !    Note the values in timeStartEnd are the actual time (aka value of aryX(i)), NOT the bin number.
  !  lowerCutoffY : (OPTIONAL) The bin with Y equal to or smaller than this value is ignored (The mask is set to be .false.).  This value has to be non-negative.
  !  lowerNbinRatio : (OPTIONAL) Passed to findOptimumY(); In search if the ratio of number of bin currently used to the initial bin number drops below this thoreshold, the calculation (aka search) will stop in findOptimumY().
  !  bkgDiagn : (OPTIONAL) Passed to findOptimumY() as it is; the diagnostic outputs from the routine will be stored.
  !  status : (OPTIONAL) Passed to findOptimumY() as it is; if this is not given and if the search for the minimum fails, then Error is raised.
  ! 

    implicit none

    real(single), intent(in) :: aryX(:), aryY(:)
    real(single), intent(out) :: optimumRate
    integer(int32),          intent(out), optional :: optimumSize, maskSize
    type(InfoFracexpT),      intent(in),  optional :: fracexp
    type(ArgTimesStartEndT), intent(in),  optional :: timeStartEnd
    real(single),            intent(in),  optional :: lowerCutoffY, lowerNbinRatio
    integer,                 intent(out), optional :: status
    type(BkgoptrateDiagnosisT), intent(out), optional :: bkgDiagn
    logical,                 intent(in),  optional :: doDump

    real(single), dimension(size(aryX)) :: tsCnt
    logical :: mask(size(aryX))
    type(InfoFracexpT) :: localFracexp
    type(ArgTimesStartEndT) :: localTimesStartEnd
    real(single) :: localLowerCutoffY, localLowerNbinRatio !, prevX
    integer(int32) :: nBinIn, i, si

    !-------------------------------------------------------------------
    ! Start

    if (present(optimumSize)) optimumSize = -1
    if (present(maskSize))    maskSize    = -1
    if (present(fracexp))      localFracexp = fracexp
    if (.not. allocated(localFracexp%ary)) allocate(localFracexp%ary(size(aryX)))
    if (present(timeStartEnd)) localTimesStartEnd = timeStartEnd
    localLowerCutoffY = 0.0
    if (present(lowerCutoffY)) localLowerCutoffY = lowerCutoffY
    if (localLowerCutoffY < 0) then
    write(errStr, '(a,es14.8,a)') "lowerCutoffY specified is negative (" &
       , lowerCutoffY &
       , ").  It is reset to zero in calcThresholdFromAny().  Contact the code developer."
      call warning('negativeLowerCutoffY')
    end if
    localLowerNbinRatio = 0.05
    if (present(lowerNbinRatio)) localLowerNbinRatio = lowerNbinRatio

    nBinIn = size(aryX)
    if (nBinIn .ne. size(aryY)) then
      call error('differentSizeOfAryTimeAndRate')
    else if (localFracexp%doUse .and. (nBinIn .ne. size(localFracexp%ary))) then
      call error('differentSizeOfAryTimeAndFracexp')
    end if

    ! Initialisation
    tsCnt = aryY
    ! if (localFracexp%doUse .and. (trim(localFracexp%style) == 'calc')) then
    !   where (localFracexp%ary ==   0) tsCnt = 0
    !   where (localFracexp%ary .ne. 0) tsCnt = aryY / localFracexp%ary
    ! else
    !   tsCnt = aryY
    ! end if

    mask(:) = .true.
    ! prevX = aryX(1) - 1	! Assuming (size(aryX) > 0)
    !! Set the initial value for prevX, which can be arbitrary as long as smaller than aryX(1).

    ! Adjusting the mask array.
    do i=1, nBinIn 
      ! Most of the IF clauses should be located ideally out of the loop
      ! for the sake of the run-time speed.  However they should not be
      ! the bottle neck, so I leave them as they are for readability.

      ! if (prevX >= aryX(i)) then
      !   write(errStr, '(a, i10, a, es16.6, es16.6)') &
      !      'Time bin in the input time series is not in order of time at Bin=' &
      !      , i, ' times:', prevX,  aryX(i)
      !   call error('notSortedTimeBin')
      ! end if

      ! prevX = aryX(i)

      ! if (aryY(i) <= 0) then
      if (.not. (aryY(i) > localLowerCutoffY)) then	! If it is below the threshold (maybe just negative or zero), or NaN.
        mask(i) = .false.
        cycle
      end if

      if (localTimesStartEnd%withStart .and. (aryX(i) < localTimesStartEnd%start)) then
        mask(i) = .false.
        cycle
      end if

      if (localFracexp%doUse .and. (trim(localFracexp%style) == 'threshold')) then
        if (localFracexp%ary(i) < localFracexp%lowerThreshold) then
          mask(i) = .false.
          cycle
        end if
      end if

      if (localTimesStartEnd%withEnd   .and. (aryX(i) > localTimesStartEnd%end)) then
        mask(i) = .false.
        cycle
      end if
    end do		! do i=1, nBinIn 


    si = count(mask)
    if (present(maskSize)) maskSize = si
    if (si <= 0) then
      write(errStr, '(a)') 'After filtering with various thresholds &
         &(GTI, Start/End-Times, Lower-Cutoff-Count, Min-Time-Ratio), &
         &no significant bin is left.'
      if (present(status)) then
        status = -4
        call warning('noSignificantBin')
        return
      else
        call error('allMaskFalse') 
      end if
    end if

    write(errStr, '(a,i8,a)') "Mask file has been successfully generated (size="&
       , si, ").  Proceed to calculate."
    call message(AppMsg, Noisy)

    if (localFracexp%doUse .and. (trim(localFracexp%style) == 'calc')) then
      call findOptimumY(tsCnt, optimumRate &
         , optimumSize=optimumSize, aryFracexp=real(localFracexp%ary, single), mask=mask &
         , lowerNbinRatio=localLowerNbinRatio, status=status &
         , bkgDiagn=bkgDiagn, doDump=doDump)	! Raise error if goes wrong and if status not given.
    else
      call findOptimumY(tsCnt, optimumRate &
         , optimumSize=optimumSize, mask=mask &
         , lowerNbinRatio=localLowerNbinRatio, status=status &
         , bkgDiagn=bkgDiagn, doDump=doDump)	! Raise error if goes wrong and if status not given.
    end if



    deallocate(localFracexp%ary)

  end subroutine calcThresholdFromAny


  !---------------------------------------------------------------------
  subroutine calcThresholdFromRate(aryX, aryY, optimumRate &
     , optimumSize, maskSize, fracexp, timeStartEnd &
     , lowerCutoffY, lowerNbinRatio, status, bkgDiagn)
  !---------------------------------------------------------------------
  ! Description:
  !  Wrapper for calcThresholdFromAny().

    implicit none

    real(single), intent(in) :: aryX(:), aryY(:)
    real(single), intent(out) :: optimumRate
    integer(int32),          intent(out), optional :: optimumSize, maskSize
    type(InfoFracexpT),      intent(in),  optional :: fracexp
    type(ArgTimesStartEndT), intent(in),  optional :: timeStartEnd
    real(single),            intent(in),  optional :: lowerCutoffY, lowerNbinRatio
    integer,                 intent(out), optional :: status
    type(BkgoptrateDiagnosisT), intent(out), optional :: bkgDiagn

    real(single) :: aryCnt(size(aryX))
    type(InfoFracexpT) :: localFracexp

    !-------------------------------------------------------------------
    ! Start

    if (present(fracexp)) localFracexp = fracexp
    aryCnt = aryY

    ! Calculating aryCnt (counts for each time bin), based on fracexp given.
    if (localFracexp%doUse) then
      if (trim(localFracexp%style) == 'calc') then
        aryCnt = 0.0
        where((aryY > 0.0) .and. (localFracexp%ary > 0.0)) aryCnt = aryY * localFracexp%ary
      end if
    end if

    call calcThresholdFromAny(aryCnt, aryY, optimumRate &
       , optimumSize=optimumSize, maskSize=maskSize &
       , fracexp=fracexp, timeStartEnd=timeStartEnd &
       , lowerCutoffY=lowerCutoffY, lowerNbinRatio=lowerNbinRatio &
       , status=status, bkgDiagn=bkgDiagn)
    ! This is not the most efficient, as tsRate, which is aryX(:) here,
    ! will be calculated again in the later stage, subroutine findOptimumY().
    ! However the overhead in practice is probably small enough.

  end subroutine calcThresholdFromRate


  !---------------------------------------------------------------------
  subroutine calcThresholdFromCount(aryX, aryY, optimumRate &
     , optimumSize, maskSize, fracexp, timeStartEnd &
     , lowerCutoffY, lowerNbinRatio, status, bkgDiagn)
  !---------------------------------------------------------------------
  ! Description:
  !  Wrapper for calcThresholdFromAny().

    implicit none

    real(single), intent(in) :: aryX(:), aryY(:)
    real(single), intent(out) :: optimumRate
    integer(int32),          intent(out), optional :: optimumSize, maskSize
    type(InfoFracexpT),      intent(in),  optional :: fracexp
    type(ArgTimesStartEndT), intent(in),  optional :: timeStartEnd
    real(single),            intent(in),  optional :: lowerCutoffY, lowerNbinRatio
    integer,                 intent(out), optional :: status
    type(BkgoptrateDiagnosisT), intent(out), optional :: bkgDiagn

    !-------------------------------------------------------------------
    ! Start

    call calcThresholdFromAny(aryX, aryY, optimumRate &
       , optimumSize=optimumSize, maskSize=maskSize &
       , fracexp=fracexp, timeStartEnd=timeStartEnd &
       , lowerCutoffY=lowerCutoffY, lowerNbinRatio=lowerNbinRatio &
       , status=status, bkgDiagn=bkgDiagn)

  end subroutine calcThresholdFromCount


end module bkgoptrate_aux
