! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!
! Module: bkgoptrate_mod
!
! Authors: Masaaki Sakano (SSC/LUX)
!          
! Description: Output the best RATE to maximise the statistics, based on the given background time-series.
! 
! Comments: 
!  * Test with a step-function type time-series.
!   This method should work well, however epsfilt, which uses
!   the Gaussian fitting, may not, because the histogram will be
!   a double-peak function and be far from Gaussian.
! 
! $Id: module.f90,v 1.5 1998/01/13 08:39:15 gvacanti Exp $
!- 
module bkgoptrate_mod

! You should write the module name in the 'use' statement, the package name
! (and version number) in the DEPEND file, and the * from the lib*.a name in
! <package>/src/Makefile.

  use dal
  ! use cal
  use param
  use errorHandling

  use bkgoptrate_aux

  !! in ssclib/
  ! use array_utils	! for smallest/largestTrueIndex
  ! use constants
  use dal_aux		! for readColToFixed, writeColData
  ! use ssc_misc	! for intToStr
  ! use minimizations	! for riddersMethod

  implicit none 

  private 

  ! Command-line parameters:

  character(256) ::&
    argBkgTsSetTabName = '',&
    argTsStyle = '',&
    argGtiStyle = '',&
    argColNameX = '',& 
    argColNameY = '',&
    argColNameFracexp = '',&
    argFracexpStyle = '',&
    argSnSetTabName = ''

  real(single) ::&
    argStartTime = 0.0,&
    argEndTime   = 0.0,&
    argLowerCutoffCount = 0.0,&
    argMinTimeRatio = 0.05	! 5%

  logical ::&
    argWithStartTime = .false.,&
    argWithEndTime   = .false.,&
    argWithLowerCutoffCount = .false.,&
    argWithMinTimeRatio = .false.,&
    argDoInsertKwds = .false.,&
    argDoOutputSnTab = .false.,&
    argIgnoreLowCntTail = .false.

  ! Derived from the Command-line parameters:
  character(256) ::&
    BkgTsSetName = '',&
    BkgTsTabName = '',&
    SnSetName = '',&
    SnTabName = ''
  character(64) ::&
    TsColNameX = '',&
    TsColNameY = ''
  character(20) ::&
    TsStyle = '',&
    GtiStyle = ''

  ! The followings must be SAVEd apparently.  Othewise compiling would fail.
  type(InfoFracexpT), save :: InfoFracexp		! defined in bkgoptrate_aux
  type(ArgTimesStartEndT), save :: TimesStartEnd	! defined in bkgoptrate_aux

  ! Other globals:

  public bkgoptrate

contains 
 
  !---------------------------------------------------------------------
  subroutine bkgoptrate
  !---------------------------------------------------------------------

    implicit none
    type(DataSetT) :: set
    type(TableT) :: tab
    integer(int32) :: numRows
    ! type(ColumnT) :: colX, colY, colFracExp
    real(single), allocatable :: aryX(:), aryY(:)
    real(single) :: optimumValue
    integer(int32) :: optimumSize, maskSize
    integer :: subStatus
    type(BkgoptrateDiagnosisT) :: bkgDiagn	! type defined in bkgoptrate_aux

logical, parameter :: localTest = .true.

    !-------------------------------------------------------------------
    ! Start

    call readParameters

    !...................................................................
    ! Read the data.

    write(errStr, '(a)') "Read Time-series from (" &
       //trim(BkgTsSetName)//':'//trim(BkgTsTabName)//")."
    call message(AppMsg, Noisy)

    set = dataSet(trim(BkgTsSetName), READ)
    tab = table(set, trim(BkgTsTabName), READ)
    numRows = numberOfRows(tab)
    allocate(aryX(numRows), aryY(numRows))

    call readColDataToFixed(tab, trim(TsColNameX), aryX)
    call readColDataToFixed(tab, trim(TsColNameY), aryY)

    if (InfoFracexp%doUse) then
      allocate(InfoFracexp%ary(numRows))
      call readColDataToFixed(tab, trim(InfoFracexp%colName), InfoFracexp%ary)
    end if

    call release(set)

    write(errStr, '(a)') "Calculating the output threshold..."
    call message(AppMsg, Noisy)

    select case(trim(TsStyle))
    case('rate')
      ! if (argDoOutputSnTab) then
      call calcThresholdFromRate( aryX, aryY, optimumValue &
         , optimumSize=optimumSize, maskSize=maskSize &
         , fracexp=InfoFracexp, timeStartEnd=TimesStartEnd &
         , lowerCutoffY=argLowerCutoffCount, lowerNbinRatio=argMinTimeRatio &
         , status=subStatus, bkgDiagn=bkgDiagn)

    case('count')
      call calcThresholdFromCount(aryX, aryY, optimumValue &
         , optimumSize=optimumSize, maskSize=maskSize &
         , fracexp=InfoFracexp, timeStartEnd=TimesStartEnd &
         , lowerCutoffY=argLowerCutoffCount, lowerNbinRatio=argMinTimeRatio &
         , status=subStatus, bkgDiagn=bkgDiagn)

    case default
call error('badTsStyle',"Contact the code developer.")	! Should have been trapped in SUBROUTINE raedParameter().
    end select	! select case(trim(TsStyle))


    if (subStatus < 0) then
      if (subStatus <= -4) then
        call error('allMaskFalse', 'After filtering with various thresholds &
         &(GTI, Start/End-Times, Lower-Cutoff-Count, Min-Time-Ratio), &
         &no significant bin is left.  Or, possibly all the bins &
         &in the input time-series are uniformly zero.')
      else if (argIgnoreLowCntTail) then
        if (subStatus <= -3) then
          call error('noThresholdFound' &
             , 'The smaller number of bins are used, better the statistics will'&
             //' be in the given condition, and no local maximum for'&
             //' the signal-to-noise ratio is found, either.')
        else
          write(errStr, '(a)') 'The bin(s) with the lowest count in the input file &
             &has the maximum S/N.  &
             &Therefore, with ignorelowcnttail=yes specified, those lowest &
             &bins are ignored and the local maximum at the bin at &
             &a higher count is regarded as the point of the best statistics.'
          call warning('lowestBinsHaveBestSN')
        end if
      else
          call error('noThresholdFound' &
             , 'The smaller number of bins are used, better the statistics will'&
             //' be in the given condition.  You may want to specify'&
             //' ignorelowcnttail=yes to ignore that lower tail.')
      end if
    else
      ! subStatus == 0 (Normal),
      !           or 1 (The lowest-count bins have the maximum S/N, but occupies the significant amount in the input lightcurve)
    end if

    !-------- Output (STDOUT) --------
    write(errStr, '(i10,a,i10,a,i10,a)') optimumSize &
       , " bins are found to be below the threshold, out of ", maskSize &
       , " bins after filtering is applied, out of ", size(aryX) &
       , " bins in the original input."
    call message(AppMsg, Noisy)

    write(errStr, '(a)') "The best Y-value is "
    call message(AppMsg, Verbose)

    write(*,'(es15.8)') optimumValue 

    !-------- Output (Insert Keywords in the input FITS) --------
    if (argDoInsertKwds) then
      set = dataSet(trim(BkgTsSetName), MODIFY)
      tab = table(set, trim(BkgTsTabName), MODIFY)

      call setAttribute(tab, 'BGOPTCUT', optimumValue &
         , comment="Optimum threshold of the BKG by bkgoptrate.")
      call message(AppMsg, Noisy, 'BGOPTCUT is written in the input file.')
      call setAttribute(tab, 'FRACTLFT', real(optimumSize)/maskSize &
         , comment="Fraction of time bins used if BGOPTCUT applied.")
      write(errStr, '(a,es13.6,a)') 'FRACBLOS =(' &
         , real(optimumSize)/maskSize &
         , ') is written in the input file.'
      call message(AppMsg, Noisy)

      call release(set)
    end if

    !-------- Output (Diagnostic table FITS) --------

    if (argDoOutputSnTab) then
      call message(AppMsg, Verbose, 'Writing the diagnostic table FITS: '&
         //trim(SnSetName)//':'//trim(SnTabName))
      set = dataSet(trim(SnSetName), CREATE)	! nb. Clobber setting by the user may prevent this.
      tab = addTable(set, trim(SnTabName), size(bkgDiagn%aryBRawCut))

      select case(trim(TsStyle))
      case('rate')
        call writeColData(tab, 'BKGRATECUT', bkgDiagn%aryBRateCut, 'counts/unitTime', 'Background threshold')
        call writeColData(tab, 'BKGRAWCUT', bkgDiagn%aryBRawCut, 'counts/unitTime', 'Background threshold')
        call writeColData(tab, 'BKGRATE',   bkgDiagn%aryRate, 'counts/unitTime', 'Exact background value')
      case('count')
        call writeColData(tab, 'BKGRATECUT', bkgDiagn%aryBRateCut, 'counts/unitTime', 'Background threshold')
        call writeColData(tab, 'BKGRAWCUT', bkgDiagn%aryBRawCut, 'counts', 'Background threshold')
        call writeColData(tab, 'BKGRATE',   bkgDiagn%aryRate, 'counts/unitTime', 'Exact background value')
      case default
        call error('badTsStyle',"Contact the code developer.")	! Should have been trapped in SUBROUTINE raedParameter().
      end select	! select case(trim(TsStyle))

      ! call writeColData(tab, 'S/N', bkgDiagn%arySN, '', 'Signal-to-Noise ratio')	! 'S/N' can not be read by fv due to the forward slash.
      call writeColData(tab, 'SN_RATIO', bkgDiagn%arySN, '', 'Signal-to-Noise ratio')
      call writeColData(tab, 'N_BIN', bkgDiagn%aryNBin, '', 'Number of active bins used')

      ! Adding some header keywords.
      call setAttribute(tab, 'BGOPTCUT', optimumValue &
         , comment="Optimum threshold of the BKG by bkgoptrate.")
      call setAttribute(tab, 'FRACTLFT', real(optimumSize)/maskSize &
         , comment="Fraction of time bins used if BGOPTCUT applied.")
      call setAttribute(tab, 'NBINOPT', optimumSize &
         , comment="Number of bins if BGOPTCUT applied.")
      call setAttribute(tab, 'NBINAFTF', maskSize &
         , comment="Number of bins after filtering is applied.")
      call setAttribute(tab, 'NBINORIG', size(aryX) &
         , comment="Number of original bins.")

      call release(set)
      call message(AppMsg, Verbose, 'Finished writing the diagnostic table FITS.')
    end if

    if (InfoFracexp%doUse) then
      deallocate(InfoFracexp%ary)
    end if

  end subroutine bkgoptrate


  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none

    type(DataSetT) :: tsSet
    type(TableT) :: tsTab

    logical :: noColonFound

    !-------------------------------------------------------------------
    ! Start

    argBkgTsSetTabName = stringParameter("tssettabname")
    noColonFound = .false.
    call splitSetTabName(argBkgTsSetTabName, BkgTsSetName, BkgTsTabName &
       , noColonFound=noColonFound)	! in ssclib/dal_aux

    tsSet = dataSet(trim(BkgTsSetName), READ)
    tsTab = table(tsSet, trim(BkgTsTabName))

    argTsStyle = stringParameter("tsstyle")
    TsStyle = trim(argTsStyle)
    argColNameX = stringParameter("xcol")	! TIME (at the Centre) in default
    argColNameY = stringParameter("ycol")

    argGtiStyle = stringParameter("gtistyle")
    argFracexpStyle = stringParameter("fracexpstyle")
    argColNameFracexp = stringParameter("fracexpcol")

    argWithStartTime = booleanParameter("withstarttime")
    if (argWithStartTime) argStartTime = realParameter("starttime")
    argWithEndTime = booleanParameter("withendtime")
    if (argWithEndTime)   argEndTime   = realParameter("endtime")

    argWithLowerCutoffCount = booleanParameter("withlowercutoffcount")
    if (argWithLowerCutoffCount) argLowerCutoffCount = realParameter("lowercutoffcount")
    argWithMinTimeRatio = booleanParameter("withmintimeratio")
    if (argWithMinTimeRatio) argMinTimeRatio = realParameter("mintimeratio")
    argIgnoreLowCntTail = booleanParameter("ignorelowcnttail")

    argDoInsertKwds  = booleanParameter("doinsertkwds")

    argDoOutputSnTab = booleanParameter("dooutputsntab")
    if (argDoOutputSnTab) then
      argSnSetTabName = stringParameter("snsettabname")
      noColonFound = .false.
      call splitSetTabName(argSnSetTabName, SnSetName, SnTabName &
         , noColonFound=noColonFound)	! in ssclib/dal_aux
    end if

    ! Determine TsStyle if argTsStyle=='auto'
    if (trim(argTsStyle) == 'auto') then
      if (     hasColumn(tsTab, 'RATE')) then
        TsStyle = 'rate'
        TsColNameY = 'RATE'
      else if (hasColumn(tsTab, 'rate')) then
        TsStyle = 'rate'
        TsColNameY = 'rate'
      else if (hasColumn(tsTab, 'Rate')) then
        TsStyle = 'rate'
        TsColNameY = 'Rate'
      else if (hasColumn(tsTab, 'COUNT')) then
        TsStyle = 'count'
        TsColNameY = 'COUNT'
      else if (hasColumn(tsTab, 'count')) then
        TsStyle = 'count'
        TsColNameY = 'count'
      else if (hasColumn(tsTab, 'Count')) then
        TsStyle = 'count'
        TsColNameY = 'Count'
      else if (hasColumn(tsTab, 'COUNTS')) then
        TsStyle = 'count'
        TsColNameY = 'COUNTS'
      else if (hasColumn(tsTab, 'counts')) then
        TsStyle = 'count'
        TsColNameY = 'counts'
      else if (hasColumn(tsTab, 'Counts')) then
        TsStyle = 'count'
        TsColNameY = 'Counts'
      else
call error('FailedInTsStyleOfAuto' &
, "Neither 'RATE' nor 'COUNTS?' is found as a column name &
&hence the --tsstyle=auto parameter does not work.")
      end if

      call message(AppMsg, Verbose, 'tsstyle is determined to be (' &
         //trim(TsStyle)//') in tsstyle=auto with the column name to read of (' &
         //trim(TsColNameY)//').')
    end if


    ! Determine TsColNameX/Y based on --x/ycol.

    if (hasColumn(tsTab, trim(argColNameX))) then
      TsColNameX = trim(argColNameX)
    else
call error('NotExistCountColName' &
, "The specified column ("//trim(argColNameY)//") in (" &
//trim(BkgTsSetName)//":"//trim(BkgTsTabName)//") does not exist.")
    end if

    select case(trim(TsStyle))
    case('rate')
      if (trim(argColNameY) == 'default') then
        if (     hasColumn(tsTab, 'RATE')) then
          TsColNameY = 'RATE'
        else if (hasColumn(tsTab, 'rate')) then
          TsColNameY = 'rate'
        else if (hasColumn(tsTab, 'Rate')) then
          TsColNameY = 'Rate'
        else
call error('FailedToFindRateColName' &
, "Failed to find the 'RATE' column in (" &
//trim(BkgTsSetName)//":"//trim(BkgTsTabName)//").")
        end if

        call message(AppMsg, Verbose, 'ycol is determined to be (' &
           //trim(TsColNameY)//') in ycol=default.')

      else if (hasColumn(tsTab, trim(argColNameY))) then
        TsColNameY = trim(argColNameY)
      else
call error('NotExistRateColName' &
, "The specified column ("//trim(argColNameY)//") does not exist in (" &
//trim(BkgTsSetName)//":"//trim(BkgTsTabName)//").")
      end if

    case('count')
      if (trim(argColNameY) == 'default') then
        if (     hasColumn(tsTab, 'COUNT')) then
          TsColNameY = 'COUNT'
        else if (hasColumn(tsTab, 'count')) then
          TsColNameY = 'count'
        else if (hasColumn(tsTab, 'Count')) then
          TsColNameY = 'Count'
        else if (hasColumn(tsTab, 'COUNTS')) then
          TsColNameY = 'COUNTS'
        else if (hasColumn(tsTab, 'counts')) then
          TsColNameY = 'counts'
        else if (hasColumn(tsTab, 'Counts')) then
          TsColNameY = 'Counts'
        else
call error('FailedToFindCountColName' &
, "Failed to find the 'COUNT' column in (" &
//trim(BkgTsSetName)//":"//trim(BkgTsTabName)//").")
        end if

        call message(AppMsg, Verbose, 'ycol is determined to be (' &
           //trim(TsColNameY)//') in ycol=default.')

      else if (hasColumn(tsTab, trim(argColNameY))) then
        TsColNameY = trim(argColNameY)
      else
call error('NotExistCountColName' &
, "The specified column ("//trim(argColNameY)//") does not exist in (" &
//trim(BkgTsSetName)//":"//trim(BkgTsTabName)//").")
      end if

    case default
call error('badTsStyle',"The value of the --tsstyle parameter is not recognized.")
    end select	! select case(trim(TsStyle))


    write(errStr, '("Column (",a,") is found in (",a,":",a,").")') &
       trim(TsColNameY), trim(BkgTsSetName), trim(BkgTsTabName)
    call message(AppMsg, Noisy)


    select case(trim(argGtiStyle))
    case('auto')
      GtiStyle = 'column'
    case('column')
      GtiStyle = 'column'

    case('gti')
    ! GTI in FITS extentions is generally defined for chip by chip.
    ! Therefore, in order to treat GTI properly, it is necessary
    ! to process like lccorr, which is far too complicated at best,
    ! particularly given lccorr does not work properly (as in 2010)
    ! due to the problem rooted in the DSSLib (fortran interface).
call error('badGtiStyle',"The value of the --gtistyle=gti &
&is not supported at the moment, yet.")

    case default
call error('badGtiStyle',"The value of the --gtistyle parameter &
&is not recognized.")
    end select


    ! Determine InfoFracexp%style if argFracexpStyle=='auto'
    InfoFracexp%doUse = .false.
    if (trim(argFracexpStyle) == 'auto') then
      if (trim(TsStyle) == 'rate') then
        InfoFracexp%style = 'none'
        call message(AppMsg, Verbose &
           , 'Because tsstyle==rate, FRACEXP is ignored (fracexpstyle=none) in fracexpstyle=auto.')
      else
        ! InfoFracexp%style will be either 'calc' or 'none'.
        if (     hasColumn(tsTab, trim(argColNameFracexp))) then
            InfoFracexp%style = 'calc'
            InfoFracexp%colName = trim(argColNameFracexp)
        else
          InfoFracexp%style = 'none'
        end if

        call message(AppMsg, Verbose, 'fracexpstyle is determined to be (' &
           //trim(InfoFracexp%style)//') in fracexpstyle=auto.')
      end if
    else
      InfoFracexp%style = trim(argFracexpStyle)
    end if

    select case(trim(InfoFracexp%style))
    case('calc', 'threshold')
      if (hasColumn(tsTab, trim(argColNameFracexp))) then
        InfoFracexp%colName = trim(argColNameFracexp)
        InfoFracexp%doUse = .true.

        write(errStr, '("Column (",a,") is found in (",a,":",a,").")') &
           trim(InfoFracexp%colName), trim(BkgTsSetName), trim(BkgTsTabName)
        call message(AppMsg, Noisy)

        if (trim(InfoFracexp%style) == 'threshold') then
          ! The bins, of which FRACEXP is lower than this, are not used.
          InfoFracexp%lowerThreshold = realParameter("fracexplower")
        end if

      else
call error('NotExistFracexpCol' &
, "The specified column ("//trim(argColNameFracexp)//") does not exist in (" &
//trim(BkgTsSetName)//":"//trim(BkgTsTabName)//").")
      end if

    case('none')
        InfoFracexp%doUse = .false.

    case default
call error('badFracexpStyle',"The value of the --fracexpstyle parameter &
&is not recognized.")
    end select


    call release(tsSet)

    TimesStartEnd%withStart = argWithStartTime
    TimesStartEnd%withEnd   = argWithEndTime
    TimesStartEnd%start = argStartTime
    TimesStartEnd%end   = argEndTime

  end subroutine readParameters


  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module bkgoptrate_mod
